import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator

import pixeltable.exceptions as excs


@contextmanager
def transactional_directory(folder_path: Path) -> Generator[Path, Any, Any]:
    """
    Args:
        folder_path: path to the folder we want to create

    Yields:
        A pathlib.Path to a hidden temporary folder, which can be used to accumulate changes.
        If everything succeeds, the changes are committed via an atomic move operation upon
        exiting the 'with' block (os.replace)
        If an exception occurred, no changes are visible in the original folder.

    Example:
        folder_path = pathlib.Path("path/to/folder")
        with transactional_folder(folder_path) as temp_folder:
            (temp_folder / "subfolder1").mkdir()
            (temp_folder / "subfolder2").mkdir()
    """
    if folder_path.exists():
        raise excs.Error(f'Folder {folder_path} already exists')

    tmp_folder = folder_path.parent / f'.tmp_{folder_path.name}'
    # Remove the temporary folder if it already exists, eg if the previous run crashed
    shutil.rmtree(str(tmp_folder), ignore_errors=True)
    tmp_folder.mkdir(parents=True)
    yield tmp_folder
    # If everything succeeds, `commit' the changes by moving the temporary folder
    tmp_folder.rename(folder_path)



from __future__ import annotations

import logging
from typing import Optional
from uuid import UUID

from pixeltable.env import Env
from pixeltable.metadata import schema

from .column import Column
from .globals import MediaValidation
from .table_version import TableVersion
from .table_version_handle import TableVersionHandle

_logger = logging.getLogger('pixeltable')


class TableVersionPath:
    """
    A TableVersionPath represents the sequence of TableVersions from a base table to a particular view:
    - for a base table: only includes that TableVersion
    - for a view: includes the TableVersion for the view and all its bases
    - multiple snapshots can reference the same TableVersion, but with different bases, which means that the
      graph of TableVersions is a DAG, not a tree (which is why we cannot embed the DAG into TableVersion directly)

    TableVersionPath contains all metadata needed to execute queries and updates against a particular version of a
    table/view.

    TableVersionPath supplies metadata needed for query construction (eg, column names), for which it uses a
    cached TableVersion instance.
    - when running inside a transaction, this instance is guaranteed to be validated
    - when running outside a transaction, we use an unvalidated instance in order to avoid repeated validation
      on every metadata-related method call (the instance won't stay validated, because TableVersionHandle.get()
      runs a local transaction, at the end of which the instance is again invalidated)
    - supplying metadata from an unvalidated instance is okay, because it needs to get revalidated anyway when a
      query actually runs (at which point there is a transaction context) - there is no guarantee that in between
      constructing a DataFrame and executing it, the underlying table schema hasn't changed (eg, a concurrent process
      could have dropped a column referenced in the query).
    """

    tbl_version: TableVersionHandle
    base: Optional[TableVersionPath]
    _cached_tbl_version: Optional[TableVersion]

    def __init__(self, tbl_version: TableVersionHandle, base: Optional[TableVersionPath] = None):
        assert tbl_version is not None
        self.tbl_version = tbl_version
        self.base = base
        self._cached_tbl_version = None

    @classmethod
    def from_md(cls, path: schema.TableVersionPath) -> TableVersionPath:
        assert len(path) > 0
        result: Optional[TableVersionPath] = None
        for tbl_id_str, effective_version in path[::-1]:
            tbl_id = UUID(tbl_id_str)
            result = TableVersionPath(TableVersionHandle(tbl_id, effective_version), base=result)
        return result

    def as_md(self) -> schema.TableVersionPath:
        result = [(self.tbl_version.id.hex, self.tbl_version.effective_version)]
        if self.base is not None:
            result.extend(self.base.as_md())
        return result

    def refresh_cached_md(self) -> None:
        from pixeltable.catalog import Catalog

        if Env.get().in_xact:
            # when we're running inside a transaction, we need to make sure to supply current metadata;
            # mixing stale metadata with current metadata leads to query construction failures
            # (multiple sqlalchemy Table instances for the same underlying table create corrupted From clauses)
            if self._cached_tbl_version is not None and self._cached_tbl_version.is_validated:
                # nothing to refresh
                return
        elif self._cached_tbl_version is not None:
            return

        with Catalog.get().begin_xact(tbl_id=self.tbl_version.id, for_write=False):
            self._cached_tbl_version = self.tbl_version.get()

    def clear_cached_md(self) -> None:
        self._cached_tbl_version = None
        if self.base is not None:
            self.base.clear_cached_md()

    @property
    def tbl_id(self) -> UUID:
        """Return the id of the table/view that this path represents"""
        return self.tbl_version.id

    def version(self) -> int:
        """Return the version of the table/view that this path represents"""
        self.refresh_cached_md()
        return self._cached_tbl_version.version

    def schema_version(self) -> int:
        """Return the version of the table/view that this path represents"""
        self.refresh_cached_md()
        return self._cached_tbl_version.schema_version

    def tbl_name(self) -> str:
        """Return the name of the table/view that this path represents"""
        self.refresh_cached_md()
        return self._cached_tbl_version.name

    def path_len(self) -> int:
        """Return the length of the path"""
        return 1 if self.base is None else 1 + self.base.path_len()

    def is_snapshot(self) -> bool:
        """Return True if this is a path of snapshot versions"""
        return self.tbl_version.is_snapshot

    def is_view(self) -> bool:
        self.refresh_cached_md()
        return self._cached_tbl_version.is_view

    def is_component_view(self) -> bool:
        self.refresh_cached_md()
        return self._cached_tbl_version.is_component_view

    def is_replica(self) -> bool:
        self.refresh_cached_md()
        return self._cached_tbl_version.is_replica

    def is_mutable(self) -> bool:
        self.refresh_cached_md()
        return self._cached_tbl_version.is_mutable

    def is_insertable(self) -> bool:
        self.refresh_cached_md()
        return self._cached_tbl_version.is_insertable

    def comment(self) -> str:
        self.refresh_cached_md()
        return self._cached_tbl_version.comment

    def num_retained_versions(self) -> int:
        self.refresh_cached_md()
        return self._cached_tbl_version.num_retained_versions

    def media_validation(self) -> MediaValidation:
        self.refresh_cached_md()
        return self._cached_tbl_version.media_validation

    def get_tbl_versions(self) -> list[TableVersionHandle]:
        """Return all tbl versions"""
        if self.base is None:
            return [self.tbl_version]
        return [self.tbl_version, *self.base.get_tbl_versions()]

    def get_bases(self) -> list[TableVersionHandle]:
        """Return all tbl versions"""
        if self.base is None:
            return []
        return self.base.get_tbl_versions()

    def find_tbl_version(self, id: UUID) -> Optional[TableVersionHandle]:
        """Return the matching TableVersion in the chain of TableVersions, starting with this one"""
        if self.tbl_version.id == id:
            return self.tbl_version
        if self.base is None:
            return None
        return self.base.find_tbl_version(id)

    def columns(self) -> list[Column]:
        """Return all user columns visible in this tbl version path, including columns from bases"""
        self.refresh_cached_md()
        result = list(self._cached_tbl_version.cols_by_name.values())
        if self.base is not None and self._cached_tbl_version.include_base_columns:
            base_cols = self.base.columns()
            # we only include base columns that don't conflict with one of our column names
            result.extend(c for c in base_cols if c.name not in self._cached_tbl_version.cols_by_name)
        return result

    def cols_by_name(self) -> dict[str, Column]:
        """Return a dict of all user columns visible in this tbl version path, including columns from bases"""
        cols = self.columns()
        return {col.name: col for col in cols}

    def cols_by_id(self) -> dict[int, Column]:
        """Return a dict of all user columns visible in this tbl version path, including columns from bases"""
        cols = self.columns()
        return {col.id: col for col in cols}

    def get_column(self, name: str, include_bases: Optional[bool] = None) -> Optional[Column]:
        """Return the column with the given name, or None if not found"""
        self.refresh_cached_md()
        col = self._cached_tbl_version.cols_by_name.get(name)
        if col is not None:
            return col
        elif self.base is not None and (include_bases or self._cached_tbl_version.include_base_columns):
            return self.base.get_column(name)
        else:
            return None

    def get_column_by_id(self, tbl_id: UUID, col_id: int) -> Optional[Column]:
        """Return the column for the given tbl/col id"""
        self.refresh_cached_md()
        if self.tbl_version.id == tbl_id:
            assert col_id in self._cached_tbl_version.cols_by_id
            return self._cached_tbl_version.cols_by_id[col_id]
        elif self.base is not None:
            return self.base.get_column_by_id(tbl_id, col_id)
        else:
            return None

    def has_column(self, col: Column, include_bases: bool = True) -> bool:
        """Return True if this table has the given column."""
        self.refresh_cached_md()
        assert col.tbl is not None
        if (
            col.tbl.id == self.tbl_version.id
            and col.tbl.effective_version == self.tbl_version.effective_version
            and col.id in self._cached_tbl_version.cols_by_id
        ):
            # the column is visible in this table version
            return True
        elif self.base is not None and include_bases:
            return self.base.has_column(col)
        else:
            return False

    def as_dict(self) -> dict:
        return {
            'tbl_version': self.tbl_version.as_dict(),
            'base': self.base.as_dict() if self.base is not None else None,
        }

    @classmethod
    def from_dict(cls, d: dict) -> TableVersionPath:
        tbl_version = TableVersionHandle.from_dict(d['tbl_version'])
        base = TableVersionPath.from_dict(d['base']) if d['base'] is not None else None
        return cls(tbl_version, base)



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Together AI API. In order to use them, you must
first `pip install together` and configure your Together AI credentials, as described in
the [Working with Together AI](https://pixeltable.readme.io/docs/together-ai) tutorial.
"""

import base64
import io
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar

import numpy as np
import PIL.Image
import requests
import tenacity

import pixeltable as pxt
import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import env
from pixeltable.func import Batch
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import together  # type: ignore[import-untyped]


@env.register_client('together')
def _(api_key: str) -> 'together.AsyncTogether':
    import together

    return together.AsyncTogether(api_key=api_key)


def _together_client() -> 'together.AsyncTogether':
    return env.Env.get().get_client('together')


T = TypeVar('T')


def _retry(fn: Callable[..., T]) -> Callable[..., T]:
    import together

    return tenacity.retry(
        retry=tenacity.retry_if_exception_type(together.error.RateLimitError),
        wait=tenacity.wait_random_exponential(multiplier=1, max=60),
        stop=tenacity.stop_after_attempt(20),
    )(fn)


@pxt.udf(resource_pool='request-rate:together:chat')
async def completions(prompt: str, *, model: str, model_kwargs: Optional[dict[str, Any]] = None) -> dict:
    """
    Generate completions based on a given prompt using a specified model.

    Equivalent to the Together AI `completions` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/completions-1>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `chat`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        prompt: A string providing context for the model to complete.
        model: The name of the model to query.
        model_kwargs: Additional keyword arguments for the Together `completions` API.
            For details on the available parameters, see: <https://docs.together.ai/reference/completions-1>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `mistralai/Mixtral-8x7B-v0.1` to an existing Pixeltable column
        `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(response=completions(tbl.prompt, model='mistralai/Mixtral-8x7B-v0.1'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    result = await _together_client().completions.create(prompt=prompt, model=model, **model_kwargs)
    return result.dict()


@pxt.udf(resource_pool='request-rate:together:chat')
async def chat_completions(
    messages: list[dict[str, str]], *, model: str, model_kwargs: Optional[dict[str, Any]] = None
) -> dict:
    """
    Generate chat completions based on a given prompt using a specified model.

    Equivalent to the Together AI `chat/completions` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/chat-completions-1>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `chat`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        messages: A list of messages comprising the conversation so far.
        model: The name of the model to query.
        model_kwargs: Additional keyword arguments for the Together `chat/completions` API.
            For details on the available parameters, see: <https://docs.together.ai/reference/chat-completions-1>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `mistralai/Mixtral-8x7B-v0.1` to an existing Pixeltable column
        `tbl.prompt` of the table `tbl`:

        >>> messages = [{'role': 'user', 'content': tbl.prompt}]
        ... tbl.add_computed_column(response=chat_completions(messages, model='mistralai/Mixtral-8x7B-v0.1'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    result = await _together_client().chat.completions.create(messages=messages, model=model, **model_kwargs)
    return result.dict()


_embedding_dimensions_cache = {
    'togethercomputer/m2-bert-80M-2k-retrieval': 768,
    'togethercomputer/m2-bert-80M-8k-retrieval': 768,
    'togethercomputer/m2-bert-80M-32k-retrieval': 768,
    'WhereIsAI/UAE-Large-V1': 1024,
    'BAAI/bge-large-en-v1.5': 1024,
    'BAAI/bge-base-en-v1.5': 768,
    'sentence-transformers/msmarco-bert-base-dot-v5': 768,
    'bert-base-uncased': 768,
}


@pxt.udf(batch_size=32, resource_pool='request-rate:together:embeddings')
async def embeddings(input: Batch[str], *, model: str) -> Batch[pxt.Array[(None,), pxt.Float]]:
    """
    Query an embedding model for a given string of text.

    Equivalent to the Together AI `embeddings` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/embeddings-2>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `embeddings`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        input: A string providing the text for the model to embed.
        model: The name of the embedding model to use.

    Returns:
        An array representing the application of the given embedding to `input`.

    Examples:
        Add a computed column that applies the model `togethercomputer/m2-bert-80M-8k-retrieval`
        to an existing Pixeltable column `tbl.text` of the table `tbl`:

        >>> tbl.add_computed_column(response=embeddings(tbl.text, model='togethercomputer/m2-bert-80M-8k-retrieval'))
    """
    result = await _together_client().embeddings.create(input=input, model=model)
    return [np.array(data.embedding, dtype=np.float64) for data in result.data]


@embeddings.conditional_return_type
def _(model: str) -> ts.ArrayType:
    if model not in _embedding_dimensions_cache:
        # TODO: find some other way to retrieve a sample
        return ts.ArrayType((None,), dtype=ts.FloatType())
    dimensions = _embedding_dimensions_cache[model]
    return ts.ArrayType((dimensions,), dtype=ts.FloatType())


@pxt.udf(resource_pool='request-rate:together:images')
async def image_generations(
    prompt: str, *, model: str, model_kwargs: Optional[dict[str, Any]] = None
) -> PIL.Image.Image:
    """
    Generate images based on a given prompt using a specified model.

    Equivalent to the Together AI `images/generations` API endpoint.
    For additional details, see: <https://docs.together.ai/reference/post_images-generations>

    Request throttling:
    Applies the rate limit set in the config (section `together.rate_limits`, key `images`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install together`

    Args:
        prompt: A description of the desired images.
        model: The model to use for image generation.
        model_kwargs: Additional keyword args for the Together `images/generations` API.
            For details on the available parameters, see: <https://docs.together.ai/reference/post_images-generations>

    Returns:
        The generated image.

    Examples:
        Add a computed column that applies the model `stabilityai/stable-diffusion-xl-base-1.0`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(
        ...     response=image_generations(tbl.prompt, model='stabilityai/stable-diffusion-xl-base-1.0')
        ... )
    """
    if model_kwargs is None:
        model_kwargs = {}

    result = await _together_client().images.generate(prompt=prompt, model=model, **model_kwargs)
    if result.data[0].b64_json is not None:
        b64_bytes = base64.b64decode(result.data[0].b64_json)
        img = PIL.Image.open(io.BytesIO(b64_bytes))
        img.load()
        return img
    if result.data[0].url is not None:
        try:
            resp = requests.get(result.data[0].url)
            with io.BytesIO(resp.content) as fp:
                image = PIL.Image.open(fp)
                image.load()
                return image
        except Exception as exc:
            raise excs.Error('Failed to download generated image from together.ai.') from exc
    raise excs.Error('Response does not contain a generated image.')


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



# ruff: noqa: F401

from .aggregation_node import AggregationNode
from .cache_prefetch_node import CachePrefetchNode
from .component_iteration_node import ComponentIterationNode
from .data_row_batch import DataRowBatch
from .exec_context import ExecContext
from .exec_node import ExecNode
from .expr_eval import ExprEvalNode
from .in_memory_data_node import InMemoryDataNode
from .row_update_node import RowUpdateNode
from .sql_node import SqlAggregationNode, SqlJoinNode, SqlLookupNode, SqlNode, SqlSampleNode, SqlScanNode



import hashlib
import urllib.parse
import urllib.request
from pathlib import Path
from typing import Optional, Union


def print_perf_counter_delta(delta: float) -> str:
    """Prints a performance counter delta in a human-readable format.

    Args:
        delta: delta in seconds

    Returns:
        Human-readable string
    """
    if delta < 1e-6:
        return f'{delta * 1e9:.2f} ns'
    elif delta < 1e-3:
        return f'{delta * 1e6:.2f} us'
    elif delta < 1:
        return f'{delta * 1e3:.2f} ms'
    else:
        return f'{delta:.2f} s'


def sha256sum(path: Union[Path, str]) -> str:
    """
    Compute the SHA256 hash of a file.
    """
    if isinstance(path, str):
        path = Path(path)

    h = hashlib.sha256()
    with open(path, 'rb') as file:
        while chunk := file.read(h.block_size):
            h.update(chunk)

    return h.hexdigest()


def parse_local_file_path(file_or_url: str) -> Optional[Path]:
    """
    Parses a string that may be either a URL or a local file path.

    If the string is a local file path or a file-scheme URL (file://), then a Path object will be returned.
    Otherwise, None will be returned.
    """
    parsed = urllib.parse.urlparse(file_or_url)
    if len(parsed.scheme) <= 1:
        # We're using `urlparse` to help distinguish file paths from URLs. If there is no scheme, then it's a file path.
        # If there's a single-character scheme, we also interpret this as a file path; this insures that drive letters
        # on Windows pathnames are correctly handled.
        return Path(file_or_url).absolute()
    elif parsed.scheme == 'file':
        return Path(urllib.parse.unquote(urllib.request.url2pathname(parsed.path)))
    else:
        return None



from __future__ import annotations

from typing import Optional

from pixeltable.metadata import schema


class MetadataUtils:
    @classmethod
    def _diff_md(
        cls, old_md: Optional[dict[int, schema.SchemaColumn]], new_md: Optional[dict[int, schema.SchemaColumn]]
    ) -> str:
        """Return a string reporting the differences in a specific entry in two dictionaries

        Results are formatted as follows:
        - If `old_md` is `None`, returns 'Initial Version'.
        - If `old_md` and `new_md` are the same, returns an empty string.
        - If there are additions, changes, or deletions, returns a string summarizing the changes.
        """
        assert new_md is not None
        if old_md is None:
            return 'Initial Version'
        if old_md == new_md:
            return ''
        added = {k: v.name for k, v in new_md.items() if k not in old_md}
        changed = {
            k: f'{old_md[k].name!r} to {v.name!r}'
            for k, v in new_md.items()
            if k in old_md and old_md[k].name != v.name
        }
        deleted = {k: v.name for k, v in old_md.items() if k not in new_md}
        if len(added) == 0 and len(changed) == 0 and len(deleted) == 0:
            return ''
        # Format the result
        t = []
        if len(added) > 0:
            t.append('Added: ' + ', '.join(added.values()))
        if len(changed) > 0:
            t.append('Renamed: ' + ', '.join(changed.values()))
        if len(deleted) > 0:
            t.append('Deleted: ' + ', '.join(deleted.values()))
        r = ', '.join(t)
        return r

    @classmethod
    def _create_md_change_dict(
        cls, md_list: Optional[list[tuple[int, dict[int, schema.SchemaColumn]]]]
    ) -> dict[int, str]:
        """Return a dictionary of schema changes by version
        Args:
            md_list: a list of tuples, each containing a version number and a metadata dictionary.
        """
        r: dict[int, str] = {}
        if md_list is None or len(md_list) == 0:
            return r

        # Sort the list in place by version number
        md_list.sort()

        first_retrieved_version = md_list[0][0]
        if first_retrieved_version == 0:
            prev_md = None
            prev_ver = -1
            start = 0
        else:
            prev_md = md_list[0][1]
            prev_ver = first_retrieved_version
            start = 1

        for ver, curr_md in md_list[start:]:
            if ver == prev_ver:
                continue
            assert ver > prev_ver
            tf = cls._diff_md(prev_md, curr_md)
            if tf != '':
                r[ver] = tf
            prev_md = curr_md
        return r



import asyncio
import threading
from typing import Any, Coroutine, TypeVar

from pixeltable.env import Env

T = TypeVar('T')

# TODO This is a temporary hack to be able to run async UDFs in contexts that are not properly handled by the existing
#   scheduler logic (e.g., as an embedding function as part of a similarity lookup). Once the scheduler is fully
#   general, it can be removed.


def run_coroutine_synchronously(coroutine: Coroutine[Any, Any, T], timeout: float = 30) -> T:
    """
    Runs the given coroutine synchronously, even if called in the context of a running event loop.
    """
    loop = Env.get().event_loop

    if threading.current_thread() is threading.main_thread():
        return loop.run_until_complete(coroutine)
    else:
        # Not in main thread, use run_coroutine_threadsafe
        return asyncio.run_coroutine_threadsafe(coroutine, loop).result(timeout)



from __future__ import annotations

import datetime
import io
import json
import logging
import typing
from collections import deque
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
import PIL.Image

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable.catalog import Catalog
from pixeltable.utils.transactional_directory import transactional_directory

if typing.TYPE_CHECKING:
    import pyarrow as pa

    import pixeltable as pxt

_logger = logging.getLogger('pixeltable')


def _write_batch(value_batch: dict[str, deque], schema: pa.Schema, output_path: Path) -> None:
    import pyarrow as pa
    from pyarrow import parquet

    pydict = {}
    for field in schema:
        if isinstance(field.type, pa.FixedShapeTensorType):
            stacked_arr = np.stack(value_batch[field.name])
            pydict[field.name] = pa.FixedShapeTensorArray.from_numpy_ndarray(stacked_arr)
        else:
            pydict[field.name] = value_batch[field.name]

    tab = pa.Table.from_pydict(pydict, schema=schema)
    parquet.write_table(tab, str(output_path))


def export_parquet(
    table_or_df: Union[pxt.Table, pxt.DataFrame],
    parquet_path: Path,
    partition_size_bytes: int = 100_000_000,
    inline_images: bool = False,
) -> None:
    """
    Exports a dataframe's data to one or more Parquet files. Requires pyarrow to be installed.

    It additionally writes the pixeltable metadata in a json file, which would otherwise
    not be available in the parquet format.

    Args:
        table_or_df : Table or Dataframe to export.
        parquet_path : Path to directory to write the parquet files to.
        partition_size_bytes : The maximum target size for each chunk. Default 100_000_000 bytes.
        inline_images : If True, images are stored inline in the parquet file. This is useful
                        for small images, to be imported as pytorch dataset. But can be inefficient
                        for large images, and cannot be imported into pixeltable.
                        If False, will raise an error if the Dataframe has any image column.
                        Default False.
    """
    from pixeltable.utils.arrow import to_arrow_schema

    df: pxt.DataFrame
    if isinstance(table_or_df, pxt.catalog.Table):
        df = table_or_df._df()
    else:
        df = table_or_df

    type_dict = {k: v.as_dict() for k, v in df.schema.items()}
    arrow_schema = to_arrow_schema(df.schema)

    if not inline_images and any(col_type.is_image_type() for col_type in df.schema.values()):
        raise excs.Error('Cannot export Dataframe with image columns when inline_images is False')

    # store the changes atomically
    with transactional_directory(parquet_path) as temp_path:
        # dump metadata json file so we can inspect what was the source of the parquet file later on.
        json.dump(df.as_dict(), (temp_path / '.pixeltable.json').open('w'))
        json.dump(type_dict, (temp_path / '.pixeltable.column_types.json').open('w'))  # keep type metadata

        batch_num = 0
        current_value_batch: dict[str, deque] = {k: deque() for k in df.schema}
        current_byte_estimate = 0

        with Catalog.get().begin_xact(for_write=False):
            for data_row in df._exec():
                for (col_name, col_type), e in zip(df.schema.items(), df._select_list_exprs):
                    val = data_row[e.slot_idx]
                    if val is None:
                        current_value_batch[col_name].append(val)
                        continue

                    assert val is not None
                    if col_type.is_image_type():
                        # images get inlined into the parquet file
                        if data_row.file_paths is not None and data_row.file_paths[e.slot_idx] is not None:
                            # if there is a file, read directly to preserve information
                            with open(data_row.file_paths[e.slot_idx], 'rb') as f:
                                val = f.read()
                        elif isinstance(val, PIL.Image.Image):
                            # if no file available, eg. bc it is computed, convert to png
                            buf = io.BytesIO()
                            val.save(buf, format='PNG')
                            val = buf.getvalue()
                        else:
                            raise excs.Error(f'unknown image type {type(val)}')
                        length = len(val)
                    elif col_type.is_string_type():
                        length = len(val)
                    elif col_type.is_video_type() or col_type.is_audio_type():
                        if data_row.file_paths is not None and data_row.file_paths[e.slot_idx] is not None:
                            val = data_row.file_paths[e.slot_idx]
                        else:
                            raise excs.Error(f'unknown audio/video type {type(val)}')
                        length = len(val)
                    elif col_type.is_json_type():
                        val = json.dumps(val)
                        length = len(val)
                    elif col_type.is_array_type():
                        length = val.nbytes
                    elif col_type.is_int_type() or col_type.is_float_type():
                        length = 8
                    elif col_type.is_bool_type():
                        length = 1
                    elif col_type.is_date_type():
                        length = 4
                    elif col_type.is_timestamp_type():
                        val = val.astimezone(datetime.timezone.utc)
                        length = 8
                    else:
                        raise excs.Error(f'unknown type {col_type} for {col_name}')

                    current_value_batch[col_name].append(val)
                    current_byte_estimate += length
                if current_byte_estimate > partition_size_bytes:
                    assert batch_num < 100_000, 'wrote too many parquet files, unclear ordering'
                    _write_batch(current_value_batch, arrow_schema, temp_path / f'part-{batch_num:05d}.parquet')
                    batch_num += 1
                    current_value_batch = {k: deque() for k in df.schema}
                    current_byte_estimate = 0

            _write_batch(current_value_batch, arrow_schema, temp_path / f'part-{batch_num:05d}.parquet')


def import_parquet(
    table: str,
    *,
    parquet_path: str,
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    **kwargs: Any,
) -> pxt.Table:
    """Creates a new base table from a Parquet file or set of files. Requires pyarrow to be installed.

    Args:
        table: Fully qualified name of the table to import the data into.
        parquet_path: Path to an individual Parquet file or directory of Parquet files.
        schema_overrides: If specified, then for each (name, type) pair in `schema_overrides`, the column with
            name `name` will be given type `type`, instead of being inferred from the Parquet dataset. The keys in
            `schema_overrides` should be the column names of the Parquet dataset (whether or not they are valid
            Pixeltable identifiers).
        primary_key: The primary key of the table (see [`create_table()`][pixeltable.create_table]).
        kwargs: Additional arguments to pass to `create_table`.

    Returns:
        A handle to the newly created table.
    """
    value = kwargs.pop('source_format', None)
    return pxt.create_table(
        table,
        source=parquet_path,
        source_format=value,
        schema_overrides=schema_overrides,
        primary_key=primary_key,
        extra_args=kwargs,
    )



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, overload

import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import catalog

from .callable_function import CallableFunction
from .expr_template_function import ExprTemplate, ExprTemplateFunction
from .function_registry import FunctionRegistry
from .globals import validate_symbol_path
from .signature import Parameter, Signature

if TYPE_CHECKING:
    from pixeltable import exprs


# Decorator invoked without parentheses: @pxt.udf
@overload
def udf(decorated_fn: Callable) -> CallableFunction: ...


# Decorator schema invoked with parentheses: @pxt.udf(**kwargs)
@overload
def udf(
    *,
    batch_size: Optional[int] = None,
    substitute_fn: Optional[Callable] = None,
    is_method: bool = False,
    is_property: bool = False,
    resource_pool: Optional[str] = None,
    type_substitutions: Optional[Sequence[dict]] = None,
    _force_stored: bool = False,
) -> Callable[[Callable], CallableFunction]: ...


# pxt.udf() called explicitly on a Table:
@overload
def udf(
    table: catalog.Table, /, *, return_value: Any = None, description: Optional[str] = None
) -> ExprTemplateFunction: ...


def udf(*args, **kwargs):  # type: ignore[no-untyped-def]
    """A decorator to create a Function from a function definition.

    Examples:
        >>> @pxt.udf
        ... def my_function(x: int) -> int:
        ...    return x + 1
    """
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        # Decorator invoked without parentheses: @pxt.udf
        # Simply call make_function with defaults.
        return make_function(decorated_fn=args[0])

    elif len(args) == 1 and isinstance(args[0], catalog.Table):
        # pxt.udf() called explicitly on a Table
        return_value = kwargs.pop('return_value', None)
        description = kwargs.pop('description', None)
        if len(kwargs) > 0:
            raise excs.Error(f'Invalid udf kwargs: {", ".join(kwargs.keys())}')
        return from_table(args[0], return_value, description)

    else:
        # Decorator schema invoked with parentheses: @pxt.udf(**kwargs)
        # Create a decorator for the specified schema.
        batch_size = kwargs.pop('batch_size', None)
        substitute_fn = kwargs.pop('substitute_fn', None)
        is_method = kwargs.pop('is_method', None)
        is_property = kwargs.pop('is_property', None)
        resource_pool = kwargs.pop('resource_pool', None)
        type_substitutions = kwargs.pop('type_substitutions', None)
        force_stored = kwargs.pop('_force_stored', False)
        if len(kwargs) > 0:
            raise excs.Error(f'Invalid @udf decorator kwargs: {", ".join(kwargs.keys())}')
        if len(args) > 0:
            raise excs.Error('Unexpected @udf decorator arguments.')

        def decorator(decorated_fn: Callable) -> CallableFunction:
            return make_function(
                decorated_fn,
                batch_size=batch_size,
                substitute_fn=substitute_fn,
                is_method=is_method,
                is_property=is_property,
                resource_pool=resource_pool,
                type_substitutions=type_substitutions,
                force_stored=force_stored,
            )

        return decorator


def make_function(
    decorated_fn: Callable,
    return_type: Optional[ts.ColumnType] = None,
    param_types: Optional[list[ts.ColumnType]] = None,
    batch_size: Optional[int] = None,
    substitute_fn: Optional[Callable] = None,
    is_method: bool = False,
    is_property: bool = False,
    resource_pool: Optional[str] = None,
    type_substitutions: Optional[Sequence[dict]] = None,
    function_name: Optional[str] = None,
    force_stored: bool = False,
) -> CallableFunction:
    """
    Constructs a `CallableFunction` from the specified parameters.
    If `substitute_fn` is specified, then `decorated_fn`
    will be used only for its signature, with execution delegated to
    `substitute_fn`.
    """
    # Obtain function_path from decorated_fn when appropriate
    if force_stored:
        # force storing the function in the db
        function_path = None
    elif decorated_fn.__module__ != '__main__' and decorated_fn.__name__.isidentifier():
        function_path = f'{decorated_fn.__module__}.{decorated_fn.__qualname__}'
    else:
        function_path = None

    # Derive function_name, if not specified explicitly
    if function_name is None:
        function_name = decorated_fn.__name__

    # Display name to use for error messages
    errmsg_name = function_name if function_path is None else function_path

    signatures: list[Signature]
    if type_substitutions is None:
        sig = Signature.create(decorated_fn, param_types, return_type)

        # batched functions must have a batched return type
        # TODO: remove 'Python' from the error messages when we have full inference with Annotated types
        if batch_size is not None and not sig.is_batched:
            raise excs.Error(f'{errmsg_name}(): batch_size is specified; Python return type must be a `Batch`')
        if batch_size is not None and len(sig.batched_parameters) == 0:
            raise excs.Error(f'{errmsg_name}(): batch_size is specified; at least one Python parameter must be `Batch`')
        if batch_size is None and len(sig.batched_parameters) > 0:
            raise excs.Error(f'{errmsg_name}(): batched parameters in udf, but no `batch_size` given')

        if is_method and is_property:
            raise excs.Error(f'Cannot specify both `is_method` and `is_property` (in function `{function_name}`)')
        if is_property and len(sig.parameters) != 1:
            raise excs.Error(
                '`is_property=True` expects a UDF with exactly 1 parameter, but '
                f'`{function_name}` has {len(sig.parameters)}'
            )
        if (is_method or is_property) and function_path is None:
            raise excs.Error('Stored functions cannot be declared using `is_method` or `is_property`')

        signatures = [sig]
    else:
        if function_path is None:
            raise excs.Error(
                f'{errmsg_name}(): type substitutions can only be used with module UDFs (not locally defined UDFs)'
            )
        if batch_size is not None:
            raise excs.Error(f'{errmsg_name}(): type substitutions cannot be used with batched functions')
        if is_method is not None or is_property is not None:
            # TODO: Support this for `is_method`?
            raise excs.Error(f'{errmsg_name}(): type substitutions cannot be used with `is_method` or `is_property`')
        signatures = [
            Signature.create(decorated_fn, param_types, return_type, type_substitutions=subst)
            for subst in type_substitutions
        ]

    if substitute_fn is None:
        py_fn = decorated_fn
    else:
        if function_path is None:
            raise excs.Error(f'{errmsg_name}(): @udf decorator with a `substitute_fn` can only be used in a module')
        py_fn = substitute_fn

    result = CallableFunction(
        signatures=signatures,
        py_fns=[py_fn] * len(signatures),  # All signatures share the same Python function
        self_path=function_path,
        self_name=function_name,
        batch_size=batch_size,
        is_method=is_method,
        is_property=is_property,
    )
    if resource_pool is not None:
        result.resource_pool(lambda: resource_pool)

    # If this function is part of a module, register it
    if function_path is not None:
        # do the validation at the very end, so it's easier to write tests for other failure scenarios
        validate_symbol_path(function_path)
        FunctionRegistry.get().register_function(function_path, result)

    return result


@overload
def expr_udf(py_fn: Callable) -> ExprTemplateFunction: ...


@overload
def expr_udf(*, param_types: Optional[list[ts.ColumnType]] = None) -> Callable[[Callable], ExprTemplateFunction]: ...


def expr_udf(*args: Any, **kwargs: Any) -> Any:
    def make_expr_template(py_fn: Callable, param_types: Optional[list[ts.ColumnType]]) -> ExprTemplateFunction:
        from pixeltable import exprs

        if py_fn.__module__ != '__main__' and py_fn.__name__.isidentifier():
            # this is a named function in a module
            function_path = f'{py_fn.__module__}.{py_fn.__qualname__}'
        else:
            function_path = None

        # TODO: verify that the inferred return type matches that of the template
        # TODO: verify that the signature doesn't contain batched parameters

        # construct Signature from the function signature
        sig = Signature.create(py_fn=py_fn, param_types=param_types, return_type=ts.InvalidType())

        var_exprs = [exprs.Variable(param.name, param.col_type) for param in sig.parameters.values()]
        # call the function with the parameter expressions to construct an Expr with parameters
        expr = py_fn(*var_exprs)
        assert isinstance(expr, exprs.Expr)
        sig.return_type = expr.col_type
        if function_path is not None:
            validate_symbol_path(function_path)
        return ExprTemplateFunction([ExprTemplate(expr, sig)], self_path=function_path, name=py_fn.__name__)

    if len(args) == 1:
        assert len(kwargs) == 0 and callable(args[0])
        return make_expr_template(args[0], None)
    else:
        assert len(args) == 0 and len(kwargs) == 1 and 'param_types' in kwargs
        return lambda py_fn: make_expr_template(py_fn, kwargs['param_types'])


def from_table(
    tbl: catalog.Table, return_value: Optional['exprs.Expr'], description: Optional[str]
) -> ExprTemplateFunction:
    """
    Constructs an `ExprTemplateFunction` from a `Table`.

    The constructed function will have one parameter for each data column in the table, which is optional (with
    default None) if and only if its column type is nullable. The output of the function is a dict of the form
    {
        'data_col_1': Variable('data_col_1', col_type_1),
        'data_col_2': Variable('data_col_2', col_type_2),
        ...,
        'computed_col_1': computed_expr_1,
        'computed_col_2': computed_expr_2,
        ...
    }
    where the computed expressions correspond to fully substituted expressions for the computed columns of the
    table. In the substitution, ColumnRefs of data columns are replaced by Variable expressions, and ColumnRefs of
    computed columns are replaced by the (previously constructed) expressions for those columns.

    If an optional `return_value` is specified, then it is used as the return value of the function in place of
    the default dict. The same substitutions will be applied to the `return_value` expression.
    """
    from pixeltable import exprs

    ancestors = [tbl, *tbl._get_base_tables()]
    ancestors.reverse()  # We must traverse the ancestors in order from base to derived

    subst: dict[exprs.Expr, exprs.Expr] = {}
    result_dict: dict[str, exprs.Expr] = {}
    params: list[Parameter] = []

    for t in ancestors:
        for name, col in t._tbl_version.get().cols_by_name.items():
            assert name not in result_dict, f'Column name is not unique: {name}'
            if col.is_computed:
                # Computed column. Apply any existing substitutions and add the new expression to the subst dict.
                new_expr = col.value_expr.copy()
                new_expr.substitute(subst)
                subst[t[name]] = new_expr  # Substitute new_expr for ColumnRefs to this column
                result_dict[name] = new_expr
            else:
                # Data column. Include it as a parameter and add a variable expression as the subst dict.
                var = exprs.Variable(name, col.col_type)
                subst[t[name]] = var  # Substitute var for ColumnRefs to this column
                result_dict[name] = var
                # Since this is a data column, it becomes a UDF parameter.
                # If the column is nullable, then the parameter will have a default value of None.
                default_value = exprs.Literal(None) if col.col_type.nullable else None
                param = Parameter(name, col.col_type, inspect._ParameterKind.POSITIONAL_OR_KEYWORD, default_value)
                params.append(param)

    if return_value is None:
        return_value = exprs.InlineDict(result_dict)
    else:
        return_value = exprs.Expr.from_object(return_value)
        return_value = return_value.copy().substitute(subst)

    if description is None:
        # Default description is the table comment
        description = tbl._get_comment()
        if len(description) == 0:
            description = f"UDF for table '{tbl._name}'"

    # TODO: Use column comments as parameter descriptions, when we have them
    argstring = '\n'.join(f'    {param.name}: of type `{param.col_type}`' for param in params)
    docstring = f'{description}\n\nArgs:\n{argstring}'

    template = ExprTemplate(return_value, Signature(return_value.col_type, params))
    fn = ExprTemplateFunction([template], name=tbl._name)
    fn.__doc__ = docstring
    return fn



from __future__ import annotations

import abc
import logging
import sys
import warnings
from typing import Any, Iterable, Iterator, Optional, Union

import more_itertools
import psycopg
import sqlalchemy as sql
from tqdm import TqdmWarning, tqdm

from pixeltable import catalog, exceptions as excs
from pixeltable.catalog.update_status import RowCountStats
from pixeltable.env import Env
from pixeltable.exec import ExecNode
from pixeltable.metadata import schema
from pixeltable.utils.exception_handler import run_cleanup
from pixeltable.utils.media_store import MediaStore
from pixeltable.utils.sql import log_explain, log_stmt

_logger = logging.getLogger('pixeltable')


class StoreBase:
    """Base class for stored tables

    Each row has the following system columns:
    - rowid columns: one or more columns that identify a user-visible row across all versions
    - v_min: version at which the row was created
    - v_max: version at which the row was deleted (or MAX_VERSION if it's still live)
    """

    tbl_version: catalog.TableVersionHandle
    sa_md: sql.MetaData
    sa_tbl: Optional[sql.Table]
    _pk_cols: list[sql.Column]
    v_min_col: sql.Column
    v_max_col: sql.Column
    base: Optional[StoreBase]

    # In my cursory experiments this was the optimal batch size: it was an improvement over 5_000 and there was no real
    # benefit to going higher.
    # TODO: Perform more rigorous experiments with different table structures and OS environments to refine this.
    __INSERT_BATCH_SIZE = 10_000

    def __init__(self, tbl_version: catalog.TableVersion):
        self.tbl_version = catalog.TableVersionHandle(
            tbl_version.id, tbl_version.effective_version, tbl_version=tbl_version
        )
        self.sa_md = sql.MetaData()
        self.sa_tbl = None
        # We need to declare a `base` variable here, even though it's only defined for instances of `StoreView`,
        # since it's referenced by various methods of `StoreBase`
        self.base = tbl_version.base.get().store_tbl if tbl_version.base is not None else None
        # we're passing in tbl_version to avoid a circular call to TableVersionHandle.get()
        self.create_sa_tbl(tbl_version)

    def system_columns(self) -> list[sql.Column]:
        return [*self._pk_cols, self.v_max_col]

    def pk_columns(self) -> list[sql.Column]:
        return self._pk_cols

    def rowid_columns(self) -> list[sql.Column]:
        return self._pk_cols[:-1]

    @abc.abstractmethod
    def _create_rowid_columns(self) -> list[sql.Column]:
        """Create and return rowid columns"""

    def _create_system_columns(self) -> list[sql.Column]:
        """Create and return system columns"""
        rowid_cols = self._create_rowid_columns()
        self.v_min_col = sql.Column('v_min', sql.BigInteger, nullable=False)
        self.v_max_col = sql.Column(
            'v_max', sql.BigInteger, nullable=False, server_default=str(schema.Table.MAX_VERSION)
        )
        self._pk_cols = [*rowid_cols, self.v_min_col]
        return [*rowid_cols, self.v_min_col, self.v_max_col]

    def create_sa_tbl(self, tbl_version: Optional[catalog.TableVersion] = None) -> None:
        """Create self.sa_tbl from self.tbl_version."""
        if tbl_version is None:
            tbl_version = self.tbl_version.get()
        system_cols = self._create_system_columns()
        all_cols = system_cols.copy()
        for col in [c for c in tbl_version.cols if c.is_stored]:
            # re-create sql.Column for each column, regardless of whether it already has sa_col set: it was bound
            # to the last sql.Table version we created and cannot be reused
            col.create_sa_cols()
            all_cols.append(col.sa_col)
            if col.stores_cellmd:
                all_cols.append(col.sa_cellmd_col)

        if self.sa_tbl is not None:
            # if we're called in response to a schema change, we need to remove the old table first
            self.sa_md.remove(self.sa_tbl)

        idxs: list[sql.Index] = []
        # index for all system columns:
        # - base x view joins can be executed as merge joins
        # - speeds up ORDER BY rowid DESC
        # - allows filtering for a particular table version in index scan
        idx_name = f'sys_cols_idx_{tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, *system_cols))

        # v_min/v_max indices: speeds up base table scans needed to propagate a base table insert or delete
        idx_name = f'vmin_idx_{tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, self.v_min_col, postgresql_using=Env.get().dbms.version_index_type))
        idx_name = f'vmax_idx_{tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, self.v_max_col, postgresql_using=Env.get().dbms.version_index_type))

        self.sa_tbl = sql.Table(self._storage_name(), self.sa_md, *all_cols, *idxs)
        # _logger.debug(f'created sa tbl for {tbl_version.id!s} (sa_tbl={id(self.sa_tbl):x}, tv={id(tbl_version):x})')

    @abc.abstractmethod
    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        """Return predicate for rowid joins to all bases"""

    @abc.abstractmethod
    def _storage_name(self) -> str:
        """Return the name of the data store table"""

    def _move_tmp_media_file(self, file_url: Optional[str], col: catalog.Column, v_min: int) -> str:
        return MediaStore.move_tmp_media_file(file_url, self.tbl_version.id, col.id, v_min)

    def _move_tmp_media_files(
        self, table_row: list[Any], media_cols_by_sql_idx: dict[int, catalog.Column], v_min: int
    ) -> None:
        """Move tmp media files that we generated to a permanent location"""
        for n, col in media_cols_by_sql_idx.items():
            table_row[n] = self._move_tmp_media_file(table_row[n], col, v_min)

    def count(self) -> int:
        """Return the number of rows visible in self.tbl_version"""
        stmt = (
            sql.select(sql.func.count('*'))
            .select_from(self.sa_tbl)
            .where(self.v_min_col <= self.tbl_version.get().version)
            .where(self.v_max_col > self.tbl_version.get().version)
        )
        conn = Env.get().conn
        result = conn.execute(stmt).scalar_one()
        assert isinstance(result, int)
        return result

    def create(self) -> None:
        """Create If Not Exists for this table"""
        conn = Env.get().conn
        stmt = sql.schema.CreateTable(self.sa_tbl).compile(conn)
        create_stmt = str(stmt)
        if_not_exists_stmt = create_stmt.replace('CREATE TABLE', 'CREATE TABLE IF NOT EXISTS')

        # Postgres seems not to handle concurrent Create Table If Not Exists correctly, we need to ignore the various
        # errors that can occur when two connections run the same Create Table statement.
        try:
            conn.execute(sql.text(if_not_exists_stmt))
        except (sql.exc.IntegrityError, sql.exc.ProgrammingError) as e:
            Env.get().console_logger.info(f'StoreBase.create() failed with: {e}')
            if (
                isinstance(e.orig, psycopg.errors.UniqueViolation)
                and 'duplicate key value violates unique constraint "pg_type_typname_nsp_index"' in str(e.orig)
            ) or (
                isinstance(e.orig, (psycopg.errors.DuplicateObject, psycopg.errors.DuplicateTable))
                and 'already exists' in str(e.orig)
            ):
                pass
            else:
                raise

    def drop(self) -> None:
        """Drop store table"""
        conn = Env.get().conn
        self.sa_md.drop_all(bind=conn)

    def add_column(self, col: catalog.Column) -> None:
        """Add column(s) to the store-resident table based on a catalog column

        Note that a computed catalog column will require two extra columns (for the computed value and for the error
        message).
        """
        assert col.is_stored
        conn = Env.get().conn
        col_type_str = col.get_sa_col_type().compile(dialect=conn.dialect)
        s_txt = f'ALTER TABLE {self._storage_name()} ADD COLUMN {col.store_name()} {col_type_str} NULL'
        added_storage_cols = [col.store_name()]
        if col.stores_cellmd:
            cellmd_type_str = col.sa_cellmd_type().compile(dialect=conn.dialect)
            s_txt += f' , ADD COLUMN {col.cellmd_store_name()} {cellmd_type_str} DEFAULT NULL'
            added_storage_cols.append(col.cellmd_store_name())

        stmt = sql.text(s_txt)
        log_stmt(_logger, stmt)
        conn.execute(stmt)
        self.create_sa_tbl()
        _logger.info(f'Added columns {added_storage_cols} to storage table {self._storage_name()}')

    def drop_column(self, col: catalog.Column) -> None:
        """Execute Alter Table Drop Column statement"""
        s_txt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.store_name()}'
        if col.stores_cellmd:
            s_txt += f' , DROP COLUMN {col.cellmd_store_name()}'
        stmt = sql.text(s_txt)
        log_stmt(_logger, stmt)
        Env.get().conn.execute(stmt)

    def ensure_columns_exist(self, cols: Iterable[catalog.Column]) -> None:
        conn = Env.get().conn
        sql_text = f'SELECT column_name FROM information_schema.columns WHERE table_name = {self._storage_name()!r}'
        result = conn.execute(sql.text(sql_text))
        existing_cols = {row[0] for row in result}
        for col in cols:
            if col.store_name() not in existing_cols:
                self.add_column(col)

    def load_column(self, col: catalog.Column, exec_plan: ExecNode, abort_on_exc: bool) -> int:
        """Update store column of a computed column with values produced by an execution plan

        Returns:
            number of rows with exceptions
        Raises:
            sql.exc.DBAPIError if there was a SQL error during execution
            excs.Error if on_error='abort' and there was an exception during row evaluation
        """
        assert col.tbl.id == self.tbl_version.id
        num_excs = 0
        num_rows = 0
        # create temp table to store output of exec_plan, with the same primary key as the store table
        tmp_name = f'temp_{self._storage_name()}'
        tmp_pk_cols = tuple(sql.Column(col.name, col.type, primary_key=True) for col in self.pk_columns())
        tmp_val_col_sql_idx = len(tmp_pk_cols)
        tmp_val_col = sql.Column(col.sa_col.name, col.sa_col.type)
        tmp_cols = [*tmp_pk_cols, tmp_val_col]
        # add error columns if the store column records errors
        if col.stores_cellmd:
            tmp_cellmd_col = sql.Column(col.sa_cellmd_col.name, col.sa_cellmd_col.type)
            tmp_cols.append(tmp_cellmd_col)
        tmp_col_names = [col.name for col in tmp_cols]

        tmp_tbl = sql.Table(tmp_name, self.sa_md, *tmp_cols, prefixes=['TEMPORARY'])
        conn = Env.get().conn
        tmp_tbl.create(bind=conn)

        row_builder = exec_plan.row_builder

        try:
            table_rows: list[tuple[Any]] = []

            # insert rows from exec_plan into temp table
            for row_batch in exec_plan:
                num_rows += len(row_batch)
                batch_table_rows: list[tuple[Any]] = []

                for row in row_batch:
                    if abort_on_exc and row.has_exc():
                        exc = row.get_first_exc()
                        raise excs.Error(f'Error while evaluating computed column {col.name!r}:\n{exc}') from exc
                    table_row, num_row_exc = row_builder.create_table_row(row, None, row.pk)
                    if col.col_type.is_media_type():
                        table_row[tmp_val_col_sql_idx] = self._move_tmp_media_file(
                            table_row[tmp_val_col_sql_idx], col, row.pk[-1]
                        )
                    num_excs += num_row_exc
                    batch_table_rows.append(tuple(table_row))

                table_rows.extend(batch_table_rows)

                if len(table_rows) >= self.__INSERT_BATCH_SIZE:
                    self.sql_insert(tmp_tbl, tmp_col_names, table_rows)
                    table_rows.clear()

            if len(table_rows) > 0:
                self.sql_insert(tmp_tbl, tmp_col_names, table_rows)

            # update store table with values from temp table
            update_stmt = sql.update(self.sa_tbl)
            for pk_col, tmp_pk_col in zip(self.pk_columns(), tmp_pk_cols):
                update_stmt = update_stmt.where(pk_col == tmp_pk_col)
            update_stmt = update_stmt.values({col.sa_col: tmp_val_col})
            if col.stores_cellmd:
                update_stmt = update_stmt.values({col.sa_cellmd_col: tmp_cellmd_col})
            log_explain(_logger, update_stmt, conn)
            conn.execute(update_stmt)

        finally:

            def remove_tmp_tbl() -> None:
                self.sa_md.remove(tmp_tbl)
                tmp_tbl.drop(bind=conn)

            run_cleanup(remove_tmp_tbl, raise_error=True)

        return num_excs

    def insert_rows(
        self,
        exec_plan: ExecNode,
        v_min: int,
        show_progress: bool = True,
        rowids: Optional[Iterator[int]] = None,
        abort_on_exc: bool = False,
    ) -> tuple[set[int], RowCountStats]:
        """Insert rows into the store table and update the catalog table's md
        Returns:
            number of inserted rows, number of exceptions, set of column ids that have exceptions
        """
        assert v_min is not None
        # TODO: total?
        num_excs = 0
        num_rows = 0
        cols_with_excs: set[int] = set()
        progress_bar: Optional[tqdm] = None  # create this only after we started executing
        row_builder = exec_plan.row_builder

        store_col_names, media_cols_by_idx = row_builder.store_column_names()

        try:
            table_rows: list[tuple[Any]] = []
            exec_plan.open()

            for row_batch in exec_plan:
                num_rows += len(row_batch)
                batch_table_rows: list[tuple[Any]] = []

                # compute batch of rows and convert them into table rows
                for row in row_batch:
                    # if abort_on_exc == True, we need to check for media validation exceptions
                    if abort_on_exc and row.has_exc():
                        exc = row.get_first_exc()
                        raise exc

                    rowid = (next(rowids),) if rowids is not None else row.pk[:-1]
                    pk = (*rowid, v_min)
                    assert len(pk) == len(self._pk_cols)
                    table_row, num_row_exc = row_builder.create_table_row(row, cols_with_excs, pk)
                    num_excs += num_row_exc

                    if show_progress:
                        if progress_bar is None:
                            warnings.simplefilter('ignore', category=TqdmWarning)
                            progress_bar = tqdm(
                                desc=f'Inserting rows into `{self.tbl_version.get().name}`',
                                unit=' rows',
                                ncols=100,
                                file=sys.stdout,
                            )
                        progress_bar.update(1)

                    self._move_tmp_media_files(table_row, media_cols_by_idx, v_min)
                    batch_table_rows.append(tuple(table_row))

                table_rows.extend(batch_table_rows)

                # if a batch is ready for insertion into the database, insert it
                if len(table_rows) >= self.__INSERT_BATCH_SIZE:
                    self.sql_insert(self.sa_tbl, store_col_names, table_rows)
                    table_rows.clear()

            # insert any remaining rows
            if len(table_rows) > 0:
                self.sql_insert(self.sa_tbl, store_col_names, table_rows)

            if progress_bar is not None:
                progress_bar.close()
            computed_values = exec_plan.ctx.num_computed_exprs * num_rows
            row_counts = RowCountStats(ins_rows=num_rows, num_excs=num_excs, computed_values=computed_values)

            return cols_with_excs, row_counts
        finally:
            exec_plan.close()

    @classmethod
    def sql_insert(cls, sa_tbl: sql.Table, store_col_names: list[str], table_rows: list[tuple[Any]]) -> None:
        assert len(table_rows) > 0
        conn = Env.get().conn
        conn.execute(sql.insert(sa_tbl), [dict(zip(store_col_names, table_row)) for table_row in table_rows])

        # TODO: Inserting directly via psycopg delivers a small performance benefit, but is somewhat fraught due to
        #     differences in the data representation that SQLAlchemy/psycopg expect. The below code will do the
        #     insertion in psycopg and can be used if/when we decide to pursue that optimization.
        # col_names_str = ", ".join(store_col_names)
        # placeholders_str = ", ".join('%s' for _ in store_col_names)
        # stmt_text = f'INSERT INTO {self.sa_tbl.name} ({col_names_str}) VALUES ({placeholders_str})'
        # conn.exec_driver_sql(stmt_text, table_rows)

    def _versions_clause(self, versions: list[Optional[int]], match_on_vmin: bool) -> sql.ColumnElement[bool]:
        """Return filter for base versions"""
        v = versions[0]
        if v is None:
            # we're looking at live rows
            clause = sql.and_(
                self.v_min_col <= self.tbl_version.get().version, self.v_max_col == schema.Table.MAX_VERSION
            )
        else:
            # we're looking at a specific version
            clause = self.v_min_col == v if match_on_vmin else self.v_max_col == v
        if len(versions) == 1:
            return clause
        return sql.and_(clause, self.base._versions_clause(versions[1:], match_on_vmin))

    def delete_rows(
        self,
        current_version: int,
        base_versions: list[Optional[int]],
        match_on_vmin: bool,
        where_clause: Optional[sql.ColumnElement[bool]],
    ) -> int:
        """Mark rows as deleted that are live and were created prior to current_version.
        Also: populate the undo columns
        Args:
            base_versions: if non-None, join only to base rows that were created at that version,
                otherwise join to rows that are live in the base's current version (which is distinct from the
                current_version parameter)
            match_on_vmin: if True, match exact versions on v_min; if False, match on v_max
            where_clause: if not None, also apply where_clause
        Returns:
            number of deleted rows
        """
        where_clause = sql.true() if where_clause is None else where_clause
        version_clause = sql.and_(self.v_min_col < current_version, self.v_max_col == schema.Table.MAX_VERSION)
        rowid_join_clause = self._rowid_join_predicate()
        base_versions_clause = (
            sql.true() if len(base_versions) == 0 else self.base._versions_clause(base_versions, match_on_vmin)
        )
        set_clause: dict[sql.Column, Union[int, sql.Column]] = {self.v_max_col: current_version}
        for index_info in self.tbl_version.get().idxs_by_name.values():
            # copy value column to undo column
            set_clause[index_info.undo_col.sa_col] = index_info.val_col.sa_col
            # set value column to NULL
            set_clause[index_info.val_col.sa_col] = None

        stmt = (
            sql.update(self.sa_tbl)
            .values(set_clause)
            .where(where_clause)
            .where(version_clause)
            .where(rowid_join_clause)
            .where(base_versions_clause)
        )
        conn = Env.get().conn
        log_explain(_logger, stmt, conn)
        status = conn.execute(stmt)
        return status.rowcount

    def dump_rows(self, version: int, filter_view: StoreBase, filter_view_version: int) -> Iterator[dict[str, Any]]:
        filter_predicate = sql.and_(
            filter_view.v_min_col <= filter_view_version,
            filter_view.v_max_col > filter_view_version,
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns(), filter_view.rowid_columns())],
        )
        stmt = (
            sql.select('*')  # TODO: Use a more specific list of columns?
            .select_from(self.sa_tbl)
            .where(self.v_min_col <= version)
            .where(self.v_max_col > version)
            .where(sql.exists().where(filter_predicate))
        )
        conn = Env.get().conn
        _logger.debug(stmt)
        log_explain(_logger, stmt, conn)
        result = conn.execute(stmt)
        for row in result:
            yield dict(zip(result.keys(), row))

    def load_rows(self, rows: Iterable[dict[str, Any]], batch_size: int = 10_000) -> None:
        """
        When instantiating a replica, we can't rely on the usual insertion code path, which contains error handling
        and other logic that doesn't apply.
        """
        conn = Env.get().conn
        for batch in more_itertools.batched(rows, batch_size):
            conn.execute(sql.insert(self.sa_tbl), batch)


class StoreTable(StoreBase):
    def __init__(self, tbl_version: catalog.TableVersion):
        assert not tbl_version.is_view
        super().__init__(tbl_version)

    def _create_rowid_columns(self) -> list[sql.Column]:
        self.rowid_col = sql.Column('rowid', sql.BigInteger, nullable=False)
        return [self.rowid_col]

    def _storage_name(self) -> str:
        return f'tbl_{self.tbl_version.id.hex}'

    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        return sql.true()


class StoreView(StoreBase):
    def __init__(self, catalog_view: catalog.TableVersion):
        assert catalog_view.is_view
        super().__init__(catalog_view)

    def _create_rowid_columns(self) -> list[sql.Column]:
        # a view row corresponds directly to a single base row, which means it needs to duplicate its rowid columns
        self.rowid_cols = [sql.Column(c.name, c.type) for c in self.base.rowid_columns()]
        return self.rowid_cols

    def _storage_name(self) -> str:
        return f'view_{self.tbl_version.id.hex}'

    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        return sql.and_(
            self.base._rowid_join_predicate(),
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns(), self.base.rowid_columns())],
        )


class StoreComponentView(StoreView):
    """A view that stores components of its base, as produced by a ComponentIterator

    PK: now also includes pos, the position returned by the ComponentIterator for the base row identified by base_rowid
    """

    rowid_cols: list[sql.Column]
    pos_col: sql.Column
    pos_col_idx: int

    def __init__(self, catalog_view: catalog.TableVersion):
        super().__init__(catalog_view)

    def _create_rowid_columns(self) -> list[sql.Column]:
        # each base row is expanded into n view rows
        self.rowid_cols = [sql.Column(c.name, c.type) for c in self.base.rowid_columns()]
        # name of pos column: avoid collisions with bases' pos columns
        self.pos_col = sql.Column(f'pos_{len(self.rowid_cols) - 1}', sql.BigInteger, nullable=False)
        self.pos_col_idx = len(self.rowid_cols)
        self.rowid_cols.append(self.pos_col)
        return self.rowid_cols

    def create_sa_tbl(self, tbl_version: Optional[catalog.TableVersion] = None) -> None:
        if tbl_version is None:
            tbl_version = self.tbl_version.get()
        super().create_sa_tbl(tbl_version)
        # we need to fix up the 'pos' column in TableVersion
        tbl_version.cols_by_name['pos'].sa_col = self.pos_col

    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        return sql.and_(
            self.base._rowid_join_predicate(),
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns()[:-1], self.base.rowid_columns())],
        )



from typing import Any, Optional, Sequence

from pixeltable import exceptions as excs, exprs, type_system as ts

from .function import Function
from .signature import Signature


class ExprTemplate:
    """
    Encapsulates a single signature of an `ExprTemplateFunction` and its associated parameterized expression,
    along with various precomputed metadata. (This is analogous to a `Callable`-`Signature` pair in a
    `CallableFunction`.)
    """

    expr: 'exprs.Expr'
    signature: Signature
    param_exprs: dict[str, 'exprs.Variable']

    def __init__(self, expr: 'exprs.Expr', signature: Signature):
        self.expr = expr
        self.signature = signature

        self.param_exprs = {name: exprs.Variable(name, param.col_type) for name, param in signature.parameters.items()}

        # validate that all variables in the expression are parameters
        for var in expr.subexprs(expr_class=exprs.Variable):
            assert var.name in self.param_exprs, f"Variable '{var.name}' in expression is not a parameter"

        # verify default values
        self.defaults: dict[str, exprs.Literal] = {}
        for param in self.signature.parameters.values():
            if param.default is None:
                continue
            self.defaults[param.name] = param.default


class ExprTemplateFunction(Function):
    """A parameterized expression from which an executable Expr is created with a function call."""

    templates: list[ExprTemplate]
    self_name: str

    def __init__(self, templates: list[ExprTemplate], self_path: Optional[str] = None, name: Optional[str] = None):
        self.templates = templates
        self.self_name = name

        super().__init__([t.signature for t in templates], self_path=self_path)

    def _update_as_overload_resolution(self, signature_idx: int) -> None:
        self.templates = [self.templates[signature_idx]]

    @property
    def template(self) -> ExprTemplate:
        assert not self.is_polymorphic
        return self.templates[0]

    def instantiate(self, args: Sequence[Any], kwargs: dict[str, Any]) -> 'exprs.Expr':
        assert not self.is_polymorphic
        template = self.template
        bound_args = self.signature.py_signature.bind(*args, **kwargs).arguments
        # apply defaults, otherwise we might have Parameters left over
        bound_args.update(
            {param_name: default for param_name, default in template.defaults.items() if param_name not in bound_args}
        )
        result = template.expr.copy()
        arg_exprs: dict[exprs.Expr, exprs.Expr] = {}
        for param_name, arg in bound_args.items():
            param_expr = template.param_exprs[param_name]
            if not isinstance(arg, exprs.Expr):
                # TODO: use the available param_expr.col_type
                arg_expr = exprs.Expr.from_object(arg)
                if arg_expr is None:
                    raise excs.Error(f'{self.self_name}(): cannot convert argument {arg} to a Pixeltable expression')
            else:
                arg_expr = arg
            arg_exprs[param_expr] = arg_expr
        result = result.substitute(arg_exprs)
        return result

    def call_return_type(self, bound_args: dict[str, 'exprs.Expr']) -> ts.ColumnType:
        """
        The call_return_type of an ExprTemplateFunction is derived from the template expression's col_type after
        substitution (unlike for UDFs, whose call_return_type is derived from an explicitly specified
        conditional_return_type).
        """
        assert not self.is_polymorphic
        template = self.template
        with_defaults = bound_args.copy()
        with_defaults.update(
            {param_name: default for param_name, default in template.defaults.items() if param_name not in bound_args}
        )
        substituted_expr = self.template.expr.copy().substitute(
            {template.param_exprs[name]: expr for name, expr in with_defaults.items()}
        )
        return substituted_expr.col_type

    def comment(self) -> Optional[str]:
        if isinstance(self.templates[0].expr, exprs.FunctionCall):
            return self.templates[0].expr.fn.comment()
        return None

    def exec(self, args: Sequence[Any], kwargs: dict[str, Any]) -> Any:
        from pixeltable import exec

        assert not self.is_polymorphic
        expr = self.instantiate(args, kwargs)
        row_builder = exprs.RowBuilder(output_exprs=[expr], columns=[], input_exprs=[])
        row_batch = exec.DataRowBatch(tbl=None, row_builder=row_builder, num_rows=1)
        row = row_batch[0]
        row_builder.eval(row, ctx=row_builder.default_eval_ctx)
        return row[row_builder.get_output_exprs()[0].slot_idx]

    @property
    def display_name(self) -> str:
        if not self.self_name and isinstance(self.templates[0].expr, exprs.FunctionCall):
            # In the common case where the templated expression is itself a FunctionCall,
            # fall back on the display name of the underlying FunctionCall
            return self.templates[0].expr.fn.display_name
        return self.self_name

    @property
    def name(self) -> str:
        return self.self_name

    @property
    def is_async(self) -> bool:
        return False

    def __str__(self) -> str:
        return str(self.templates[0].expr)

    def _as_dict(self) -> dict:
        if self.self_path is not None:
            return super()._as_dict()
        assert not self.is_polymorphic
        assert len(self.templates) == 1
        return {'expr': self.template.expr.as_dict(), 'signature': self.signature.as_dict(), 'name': self.name}

    @classmethod
    def _from_dict(cls, d: dict) -> Function:
        if 'expr' not in d:
            return super()._from_dict(d)
        assert 'signature' in d and 'name' in d
        template = ExprTemplate(exprs.Expr.from_dict(d['expr']), Signature.from_dict(d['signature']))
        return cls([template], name=d['name'])



from typing import Generic, Iterable, Iterator, Optional, TypeVar

from .expr import Expr

T = TypeVar('T')


class ExprDict(Generic[T]):
    """
    A dictionary that maps Expr instances to values of type T.

    We cannot use dict[Expr, T] because Expr.__eq__() serves a different purpose than the default __eq__.
    """

    _data: dict[int, tuple[Expr, T]]

    def __init__(self, iterable: Optional[Iterable[tuple[Expr, T]]] = None):
        self._data = {}

        if iterable is not None:
            for key, value in iterable:
                self[key] = value

    def __setitem__(self, key: Expr, value: T) -> None:
        self._data[key.id] = (key, value)

    def __getitem__(self, key: Expr) -> T:
        return self._data[key.id][1]

    def __delitem__(self, key: Expr) -> None:
        del self._data[key.id]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[Expr]:
        return (expr for expr, _ in self._data.values())

    def __contains__(self, key: Expr) -> bool:
        return key.id in self._data

    def get(self, key: Expr, default: Optional[T] = None) -> Optional[T]:
        item = self._data.get(key.id)
        return item[1] if item is not None else default

    def clear(self) -> None:
        self._data.clear()

    def keys(self) -> Iterator[Expr]:
        return iter(self)

    def values(self) -> Iterator[T]:
        return (value for _, value in self._data.values())

    def items(self) -> Iterator[tuple[Expr, T]]:
        return ((expr, value) for expr, value in self._data.values())



import logging
from typing import Any, AsyncIterator, Optional

from pixeltable import catalog, exprs
from pixeltable.utils.media_store import MediaStore

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode

_logger = logging.getLogger('pixeltable')


class InMemoryDataNode(ExecNode):
    """
    Outputs in-memory data as a DataRowBatch of a particular table.

    Populates slots of all non-computed columns (ie, output ColumnRefs)
    - with the values provided in the input rows
    - if an input row doesn't provide a value, sets the slot to the column default
    """

    tbl: catalog.TableVersionHandle

    input_rows: list[dict[str, Any]]
    start_row_id: int
    output_rows: Optional[DataRowBatch]

    # output_exprs is declared in the superclass, but we redeclare it here with a more specific type
    output_exprs: list[exprs.ColumnRef]

    def __init__(
        self,
        tbl: catalog.TableVersionHandle,
        rows: list[dict[str, Any]],
        row_builder: exprs.RowBuilder,
        start_row_id: int,
    ):
        # we materialize the input slots
        output_exprs = list(row_builder.input_exprs)
        super().__init__(row_builder, output_exprs, [], None)
        assert tbl.get().is_insertable
        self.tbl = tbl
        self.input_rows = rows
        self.start_row_id = start_row_id
        self.output_rows = None

    def _open(self) -> None:
        """Create row batch and populate with self.input_rows"""
        user_cols_by_name = {
            col_ref.col.name: exprs.ColumnSlotIdx(col_ref.col, col_ref.slot_idx)
            for col_ref in self.output_exprs
            if col_ref.col.name is not None
        }
        output_cols_by_idx = {
            col_ref.slot_idx: exprs.ColumnSlotIdx(col_ref.col, col_ref.slot_idx) for col_ref in self.output_exprs
        }
        output_slot_idxs = {e.slot_idx for e in self.output_exprs}

        self.output_rows = DataRowBatch(self.tbl, self.row_builder, len(self.input_rows))
        for row_idx, input_row in enumerate(self.input_rows):
            # populate the output row with the values provided in the input row
            input_slot_idxs: set[int] = set()
            for col_name, val in input_row.items():
                col_info = user_cols_by_name.get(col_name)
                assert col_info is not None
                col = col_info.col
                if col.col_type.is_image_type() and isinstance(val, bytes):
                    # this is a literal media file, ie, a sequence of bytes; save it as a binary file and store the path
                    assert col.tbl.id == self.tbl.id
                    path = MediaStore.save_media_file(val, col.tbl.id, col.id, col.tbl.version)
                    self.output_rows[row_idx][col_info.slot_idx] = str(path)
                else:
                    self.output_rows[row_idx][col_info.slot_idx] = val

                input_slot_idxs.add(col_info.slot_idx)

            # set the remaining output slots to their default values (presently None)
            missing_slot_idxs = output_slot_idxs - input_slot_idxs
            for slot_idx in missing_slot_idxs:
                col_info = output_cols_by_idx.get(slot_idx)
                assert col_info is not None
                self.output_rows[row_idx][col_info.slot_idx] = None

        self.ctx.num_rows = len(self.output_rows)

    async def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        _logger.debug(f'InMemoryDataNode: created row batch with {len(self.output_rows)} output_rows')
        yield self.output_rows



"""
Pixeltable [UDF](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wraps the OpenAI Whisper library.

This UDF will cause Pixeltable to invoke the relevant model locally. In order to use it, you must
first `pip install openai-whisper`.
"""

from typing import TYPE_CHECKING, Optional, Sequence

import pixeltable as pxt
from pixeltable.env import Env

if TYPE_CHECKING:
    from whisper import Whisper  # type: ignore[import-untyped]


@pxt.udf
def transcribe(
    audio: pxt.Audio,
    *,
    model: str,
    temperature: Optional[Sequence[float]] = (0.0, 0.2, 0.4, 0.6, 0.8, 1.0),
    compression_ratio_threshold: Optional[float] = 2.4,
    logprob_threshold: Optional[float] = -1.0,
    no_speech_threshold: Optional[float] = 0.6,
    condition_on_previous_text: bool = True,
    initial_prompt: Optional[str] = None,
    word_timestamps: bool = False,
    prepend_punctuations: str = '"\'“¿([{-',
    append_punctuations: str = '"\'.。,，!！?？:：”)]}、',  # noqa: RUF001
    decode_options: Optional[dict] = None,
) -> dict:
    """
    Transcribe an audio file using Whisper.

    This UDF runs a transcription model _locally_ using the Whisper library,
    equivalent to the Whisper `transcribe` function, as described in the
    [Whisper library documentation](https://github.com/openai/whisper).

    __Requirements:__

    - `pip install openai-whisper`

    Args:
        audio: The audio file to transcribe.
        model: The name of the model to use for transcription.

    Returns:
        A dictionary containing the transcription and various other metadata.

    Examples:
        Add a computed column that applies the model `base.en` to an existing Pixeltable column `tbl.audio`
        of the table `tbl`:

        >>> tbl.add_computed_column(result=transcribe(tbl.audio, model='base.en'))
    """
    Env.get().require_package('whisper')
    Env.get().require_package('torch')
    import torch

    if decode_options is None:
        decode_options = {}
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model = _lookup_model(model, device)
    result = model.transcribe(
        audio,
        temperature=tuple(temperature),
        compression_ratio_threshold=compression_ratio_threshold,
        logprob_threshold=logprob_threshold,
        no_speech_threshold=no_speech_threshold,
        condition_on_previous_text=condition_on_previous_text,
        initial_prompt=initial_prompt,
        word_timestamps=word_timestamps,
        prepend_punctuations=prepend_punctuations,
        append_punctuations=append_punctuations,
        **decode_options,
    )
    return result


def _lookup_model(model_id: str, device: str) -> 'Whisper':
    import whisper

    key = (model_id, device)
    if key not in _model_cache:
        model = whisper.load_model(model_id, device)
        _model_cache[key] = model
    return _model_cache[key]


_model_cache: dict[tuple[str, str], 'Whisper'] = {}



from keyword import iskeyword as is_python_keyword
from typing import Any, Optional, Union

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable.catalog.globals import is_system_column_name


def normalize_pxt_col_name(name: str) -> str:
    """
    Normalizes an arbitrary DataFrame column name into a valid Pixeltable identifier by:
    - replacing any non-ascii or non-alphanumeric characters with an underscore _
    - prefixing the result with the letter 'c' if it starts with an underscore or a number
    """
    id = ''.join(ch if ch.isascii() and ch.isalnum() else '_' for ch in name)
    if id[0].isnumeric():
        id = f'c_{id}'
    elif id[0] == '_':
        id = f'c{id}'
    assert pxt.catalog.is_valid_identifier(id), id
    return id


def normalize_primary_key_parameter(primary_key: Optional[Union[str, list[str]]] = None) -> list[str]:
    if primary_key is None:
        primary_key = []
    elif isinstance(primary_key, str):
        primary_key = [primary_key]
    elif not isinstance(primary_key, list) or not all(isinstance(pk, str) for pk in primary_key):
        raise excs.Error('primary_key must be a single column name or a list of column names')
    return primary_key


def _is_usable_as_column_name(name: str, destination_schema: dict[str, Any]) -> bool:
    return not (is_system_column_name(name) or is_python_keyword(name) or name in destination_schema)


def normalize_schema_names(
    in_schema: dict[str, Any],
    primary_key: list[str],
    schema_overrides: dict[str, Any],
    require_valid_pxt_column_names: bool = False,
) -> tuple[dict[str, Any], list[str], Optional[dict[str, str]]]:
    """
    Convert all names in the input schema from source names to valid Pixeltable identifiers
    - Ensure that all names are unique.
    - Report an error if any types are missing
    - If "require_valid_pxt_column_names", report an error if any column names are not valid Pixeltable column names
    - Report an error if any primary key columns are missing
    Returns
    - A new schema with normalized column names
    - The primary key columns, mapped to the normalized names
    - A mapping from the original names to the normalized names.
    """

    # Report any untyped columns as an error
    untyped_cols = [in_name for in_name, column_type in in_schema.items() if column_type is None]
    if len(untyped_cols) > 0:
        raise excs.Error(f'Could not infer pixeltable type for column(s): {", ".join(untyped_cols)}')

    # Report any columns in `schema_overrides` that are not in the source
    extraneous_overrides = schema_overrides.keys() - in_schema.keys()
    if len(extraneous_overrides) > 0:
        raise excs.Error(
            f'Some column(s) specified in `schema_overrides` are not present '
            f'in the source: {", ".join(extraneous_overrides)}'
        )

    schema: dict[str, Any] = {}
    col_mapping: dict[str, str] = {}  # Maps column names to Pixeltable column names if needed
    for in_name, pxt_type in in_schema.items():
        pxt_name = normalize_pxt_col_name(in_name)
        # Ensure that column names are unique by appending a distinguishing suffix
        # to any collisions
        pxt_fname = pxt_name
        n = 1
        while not _is_usable_as_column_name(pxt_fname, schema):
            pxt_fname = f'{pxt_name}_{n}'
            n += 1
        schema[pxt_fname] = pxt_type
        col_mapping[in_name] = pxt_fname

    # Determine if the col_mapping is the identity mapping
    non_identity_keys = [k for k, v in col_mapping.items() if k != v]
    if len(non_identity_keys) > 0:
        if require_valid_pxt_column_names:
            raise excs.Error(
                f'Column names must be valid pixeltable identifiers. Invalid names: {", ".join(non_identity_keys)}'
            )
    else:
        col_mapping = None

    # Report any primary key columns that are not in the source as an error
    missing_pk = [pk for pk in primary_key if pk not in in_schema]
    if len(missing_pk) > 0:
        raise excs.Error(f'Primary key column(s) are not found in the source: {", ".join(missing_pk)}')

    pxt_pk = [col_mapping[pk] for pk in primary_key] if col_mapping is not None else primary_key

    return schema, pxt_pk, col_mapping



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md, convert_table_schema_version_md


@register_converter(version=21)
def _(engine: sql.engine.Engine) -> None:
    convert_table_schema_version_md(
        engine,
        table_schema_version_md_updater=__update_table_schema_version,
        schema_column_updater=__update_schema_column,
    )
    convert_table_md(engine, substitution_fn=__substitute_md)


def __update_table_schema_version(table_schema_version_md: dict) -> None:
    table_schema_version_md['media_validation'] = 'on_write'  # MediaValidation.ON_WRITE


def __update_schema_column(schema_column: dict) -> None:
    schema_column['media_validation'] = None


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    if isinstance(v, dict) and '_classname' in v and v['_classname'] == 'ColumnRef':
        if 'perform_validation' not in v:
            v['perform_validation'] = False
        return k, v
    return None



import copy
import logging
from typing import Any, Callable, Optional
from uuid import UUID

import sqlalchemy as sql

from pixeltable.metadata.schema import Function, Table, TableSchemaVersion, TableVersion

__logger = logging.getLogger('pixeltable')


def convert_table_md(
    engine: sql.engine.Engine,
    table_md_updater: Optional[Callable[[dict, UUID], None]] = None,
    column_md_updater: Optional[Callable[[dict], None]] = None,
    external_store_md_updater: Optional[Callable[[dict], None]] = None,
    substitution_fn: Optional[Callable[[Optional[str], Any], Optional[tuple[Optional[str], Any]]]] = None,
    table_modifier: Optional[Callable[[sql.Connection, UUID, dict, dict], None]] = None,
) -> None:
    """
    Converts schema.TableMd dicts based on the specified conversion functions.

    Args:
        engine: The SQLAlchemy engine.
        table_md_updater: A function that updates schema.TableMd dicts in place.
            It takes two arguments: the metadata dict (new values) and the table id.
        column_md_updater: A function that updates schema.ColumnMd dicts in place.
        external_store_md_updater: A function that updates the external store metadata in place.
        substitution_fn: A function that substitutes metadata values. If specified, all metadata will be traversed
            recursively, and `substitution_fn` will be called once for each metadata entry. If the entry appears in
            a dict as a `(k, v)` pair, then `substitution_fn(k, v)` will be called. If the entry appears in a list,
            then `substitution_fn(None, v)` will be called. If `substitution_fn` returns a tuple `(k', v')`, then
            the original entry will be replaced, and the traversal will continue with `v'`.
    """
    with engine.begin() as conn:
        # avoid a SELECT * here, which breaks when we add new columns to Table
        for row in conn.execute(sql.select(Table.id, Table.md)):
            tbl_id = row[0]
            table_md = row[1]
            assert isinstance(table_md, dict)
            updated_table_md = copy.deepcopy(table_md)
            if table_md_updater is not None:
                table_md_updater(updated_table_md, tbl_id)
            if column_md_updater is not None:
                __update_column_md(updated_table_md, column_md_updater)
            if external_store_md_updater is not None:
                __update_external_store_md(updated_table_md, external_store_md_updater)
            if substitution_fn is not None:
                updated_table_md = __substitute_md_rec(updated_table_md, substitution_fn)
            if updated_table_md != table_md:
                __logger.info(f'Updating schema for table: {tbl_id}')
                conn.execute(sql.update(Table).where(Table.id == tbl_id).values(md=updated_table_md))
            if table_modifier is not None:
                table_modifier(conn, tbl_id, table_md, updated_table_md)

        for row in conn.execute(sql.select(Function)):
            fn_id = row[0]
            function_md = row[2]
            assert isinstance(function_md, dict)
            updated_function_md = copy.deepcopy(function_md)
            if substitution_fn is not None:
                updated_function_md = __substitute_md_rec(updated_function_md, substitution_fn)
            if updated_function_md != function_md:
                __logger.info(f'Updating function: {fn_id}')
                conn.execute(sql.update(Function).where(Function.id == fn_id).values(md=updated_function_md))


def __update_column_md(table_md: dict, column_md_updater: Callable[[dict], None]) -> None:
    columns_md = table_md['column_md']
    assert isinstance(columns_md, dict)
    for column_md in columns_md.values():
        column_md_updater(column_md)


def __update_external_store_md(table_md: dict, external_store_md_updater: Callable[[dict], None]) -> None:
    stores_md = table_md['external_stores']
    assert isinstance(stores_md, list)
    for store_md in stores_md:
        external_store_md_updater(store_md)


def __substitute_md_rec(
    md: Any, substitution_fn: Callable[[Optional[str], Any], Optional[tuple[Optional[str], Any]]]
) -> Any:
    if isinstance(md, dict):
        updated_dict: dict[str, Any] = {}
        for k, v in md.items():
            assert isinstance(k, str)
            substitute = substitution_fn(k, v)
            if substitute is not None:
                updated_k, updated_v = substitute
                updated_dict[updated_k] = __substitute_md_rec(updated_v, substitution_fn)
            else:
                updated_dict[k] = __substitute_md_rec(v, substitution_fn)
        return updated_dict
    elif isinstance(md, list):
        updated_list: list[Any] = []
        for v in md:
            substitute = substitution_fn(None, v)
            if substitute is not None:
                _, updated_v = substitute
                updated_list.append(__substitute_md_rec(updated_v, substitution_fn))
            else:
                updated_list.append(__substitute_md_rec(v, substitution_fn))
        return updated_list
    else:
        return md


def convert_table_schema_version_md(
    engine: sql.engine.Engine,
    table_schema_version_md_updater: Optional[Callable[[dict], None]] = None,
    schema_column_updater: Optional[Callable[[dict], None]] = None,
) -> None:
    """
    Converts schema.TableSchemaVersionMd dicts based on the specified conversion functions.

    Args:
        engine: The SQLAlchemy engine.
        table_schema_version_md_updater: A function that updates schema.TableSchemaVersionMd dicts in place.
        schema_column_updater: A function that updates schema.SchemaColumn dicts in place.
    """
    with engine.begin() as conn:
        stmt = sql.select(TableSchemaVersion.tbl_id, TableSchemaVersion.schema_version, TableSchemaVersion.md)
        for row in conn.execute(stmt):
            tbl_id, schema_version, md = row[0], row[1], row[2]
            assert isinstance(md, dict)
            updated_md = copy.deepcopy(md)
            if table_schema_version_md_updater is not None:
                table_schema_version_md_updater(updated_md)
            if schema_column_updater is not None:
                __update_schema_column(updated_md, schema_column_updater)
            if updated_md != md:
                __logger.info(f'Updating TableSchemaVersion(tbl_id={tbl_id}, schema_version={schema_version})')
                update_stmt = (
                    sql.update(TableSchemaVersion)
                    .where(TableSchemaVersion.tbl_id == tbl_id)
                    .where(TableSchemaVersion.schema_version == schema_version)
                    .values(md=updated_md)
                )
                conn.execute(update_stmt)


def __update_schema_column(table_schema_version_md: dict, schema_column_updater: Callable[[dict], None]) -> None:
    cols = table_schema_version_md['columns']
    assert isinstance(cols, dict)
    for schema_col in cols.values():
        schema_column_updater(schema_col)


def convert_table_version_record(
    engine: sql.engine.Engine, table_version_record_updater: Optional[Callable[[TableVersion], None]]
) -> None:
    with sql.orm.Session(engine, future=True) as session:
        for record in session.query(TableVersion).all():
            table_version_record_updater(record)
        session.commit()


def convert_table_schema_version_record(
    engine: sql.engine.Engine, table_schema_version_record_updater: Optional[Callable[[TableSchemaVersion], None]]
) -> None:
    with sql.orm.Session(engine, future=True) as session:
        for record in session.query(TableSchemaVersion).all():
            table_schema_version_record_updater(record)
        session.commit()



import types
from typing import Optional

from pixeltable.func import Function

# Utilities related to the organization of the Pixeltable codebase.


def local_public_names(mod_name: str, exclude: Optional[list[str]] = None) -> list[str]:
    """
    Returns a list of all functions and submodules that are local to the specified module and are
    publicly accessible. Intended to facilitate implementation of module __dir__() methods for
    friendly tab-completion.
    """
    import importlib

    if exclude is None:
        exclude = []
    mod = importlib.import_module(mod_name)
    names = []
    for obj in mod.__dict__.values():
        if isinstance(obj, Function):
            # Pixeltable function
            names.append(obj.name)
        elif isinstance(obj, types.FunctionType):
            # Python function
            if obj.__module__ == mod.__name__ and not obj.__name__.startswith('_'):
                names.append(obj.__name__)
        elif isinstance(obj, types.ModuleType):
            # Module
            components = obj.__name__.split('.')
            if mod_name == '.'.join(components[:-1]):
                names.append(components[-1])
    return [name for name in names if name not in exclude]



from __future__ import annotations

from typing import Any, Iterable, Optional

import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

from .data_row import DataRow
from .expr import Expr
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class InPredicate(Expr):
    """Predicate corresponding to the SQL IN operator."""

    def __init__(self, lhs: Expr, value_set_literal: Optional[Iterable] = None, value_set_expr: Optional[Expr] = None):
        assert (value_set_literal is None) != (value_set_expr is None)
        if not lhs.col_type.is_scalar_type():
            raise excs.Error(f'isin(): only supported for scalar types, not {lhs.col_type}')
        super().__init__(ts.BoolType())

        self.value_list: Optional[list] = None  # only contains values of the correct type
        if value_set_expr is not None:
            if not value_set_expr.col_type.is_json_type():
                raise excs.Error(
                    f'isin(): argument must have a JSON type, but {value_set_expr} has type {value_set_expr.col_type}'
                )
            self.components = [lhs.copy(), value_set_expr.copy()]
        else:
            assert value_set_literal is not None
            self.components = [lhs.copy()]
            self.value_list = self._normalize_value_set(value_set_literal)

        self.id = self._create_id()

    @property
    def _lhs(self) -> Expr:
        return self.components[0]

    @property
    def _value_set_expr(self) -> Expr:
        assert len(self.components) == 2
        return self.components[1]

    def _normalize_value_set(self, value_set: Iterable, filter_type_mismatches: bool = True) -> list:
        if not isinstance(value_set, Iterable):
            raise excs.Error(f'isin(): argument must be an Iterable (eg, list, dict, ...), not {value_set!r}')
        value_list = list(value_set)
        if not filter_type_mismatches:
            return value_list

        # ignore elements of the wrong type
        result = []
        for val in value_list:
            try:
                self._lhs.col_type.validate_literal(val)
                result.append(val)
            except TypeError:
                pass
        return result

    def __repr__(self) -> str:
        if self.value_list is not None:
            return f'{self.components[0]}.isin({self.value_list})'
        return f'{self.components[0]}.isin({self.components[1]})'

    def _equals(self, other: InPredicate) -> bool:
        return self.value_list == other.value_list

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('value_list', self.value_list)]

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        lhs_sql_exprs = sql_elements.get(self.components[0])
        if lhs_sql_exprs is None or self.value_list is None:
            return None
        return lhs_sql_exprs.in_(self.value_list)

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        lhs_val = data_row[self._lhs.slot_idx]
        if self.value_list is not None:
            data_row[self.slot_idx] = lhs_val in self.value_list
        else:
            value_set = data_row[self._value_set_expr.slot_idx]
            value_list = self._normalize_value_set(value_set, filter_type_mismatches=False)
            data_row[self.slot_idx] = lhs_val in value_list

    def _as_dict(self) -> dict:
        return {'value_list': self.value_list, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> InPredicate:
        assert 'value_list' in d
        assert len(components) <= 2
        return cls(components[0], d['value_list'], components[1] if len(components) == 2 else None)



from __future__ import annotations

from typing import Any, Iterable, Optional, Union

import pixeltable as pxt
import pixeltable.type_system as ts
from pixeltable import exceptions as excs


def _infer_schema_from_rows(
    rows: Iterable[dict[str, Any]], schema_overrides: dict[str, ts.ColumnType], primary_key: list[str]
) -> dict[str, ts.ColumnType]:
    schema: dict[str, ts.ColumnType] = {}
    cols_with_nones: set[str] = set()

    for n, row in enumerate(rows):
        for col_name, value in row.items():
            if col_name in schema_overrides:
                # We do the insertion here; this will ensure that the column order matches the order
                # in which the column names are encountered in the input data, even if `schema_overrides`
                # is specified.
                if col_name not in schema:
                    assert isinstance(schema_overrides[col_name], ts.ColumnType)
                    schema[col_name] = schema_overrides[col_name]
            elif value is not None:
                # If `key` is not in `schema_overrides`, then we infer its type from the data.
                # The column type will always be nullable by default.
                col_type = ts.ColumnType.infer_literal_type(value, nullable=col_name not in primary_key)
                if col_type is None:
                    raise excs.Error(
                        f'Could not infer type for column `{col_name}`; the value in row {n} '
                        f'has an unsupported type: {type(value)}'
                    )
                if col_name not in schema:
                    schema[col_name] = col_type
                else:
                    supertype = schema[col_name].supertype(col_type)
                    if supertype is None:
                        raise excs.Error(
                            f'Could not infer type of column `{col_name}`; the value in row {n} '
                            f'does not match preceding type {schema[col_name]}: {value!r}\n'
                            'Consider specifying the type explicitly in `schema_overrides`.'
                        )
                    schema[col_name] = supertype
            else:
                cols_with_nones.add(col_name)

    entirely_none_cols = cols_with_nones - schema.keys()
    if len(entirely_none_cols) > 0:
        # A column can only end up in `entirely_none_cols` if it was not in `schema_overrides` and
        # was not encountered in any row with a non-None value.
        raise excs.Error(
            f'The following columns have no non-null values: {", ".join(entirely_none_cols)}\n'
            'Consider specifying the type(s) explicitly in `schema_overrides`.'
        )
    return schema


def import_rows(
    tbl_path: str,
    rows: list[dict[str, Any]],
    *,
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
) -> pxt.Table:
    """
    Creates a new base table from a list of dictionaries. The dictionaries must be of the
    form `{column_name: value, ...}`. Pixeltable will attempt to infer the schema of the table from the
    supplied data, using the most specific type that can represent all the values in a column.

    If `schema_overrides` is specified, then for each entry `(column_name, type)` in `schema_overrides`,
    Pixeltable will force the specified column to the specified type (and will not attempt any type inference
    for that column).

    All column types of the new table will be nullable unless explicitly specified as non-nullable in
    `schema_overrides`.

    Args:
        tbl_path: The qualified name of the table to create.
        rows: The list of dictionaries to import.
        schema_overrides: If specified, then columns in `schema_overrides` will be given the specified types
            as described above.
        primary_key: The primary key of the table (see [`create_table()`][pixeltable.create_table]).
        num_retained_versions: The number of retained versions of the table
            (see [`create_table()`][pixeltable.create_table]).
        comment: A comment to attach to the table (see [`create_table()`][pixeltable.create_table]).

    Returns:
        A handle to the newly created [`Table`][pixeltable.Table].
    """
    return pxt.create_table(
        tbl_path,
        source=rows,
        schema_overrides=schema_overrides,
        primary_key=primary_key,
        num_retained_versions=num_retained_versions,
        comment=comment,
    )


def import_json(
    tbl_path: str,
    filepath_or_url: str,
    *,
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    num_retained_versions: int = 10,
    comment: str = '',
    **kwargs: Any,
) -> pxt.Table:
    """
    Creates a new base table from a JSON file. This is a convenience method and is
    equivalent to calling `import_data(table_path, json.loads(file_contents, **kwargs), ...)`, where `file_contents`
    is the contents of the specified `filepath_or_url`.

    Args:
        tbl_path: The name of the table to create.
        filepath_or_url: The path or URL of the JSON file.
        schema_overrides: If specified, then columns in `schema_overrides` will be given the specified types
            (see [`import_rows()`][pixeltable.io.import_rows]).
        primary_key: The primary key of the table (see [`create_table()`][pixeltable.create_table]).
        num_retained_versions: The number of retained versions of the table
            (see [`create_table()`][pixeltable.create_table]).
        comment: A comment to attach to the table (see [`create_table()`][pixeltable.create_table]).
        kwargs: Additional keyword arguments to pass to `json.loads`.

    Returns:
        A handle to the newly created [`Table`][pixeltable.Table].
    """
    return pxt.create_table(
        tbl_path,
        source=filepath_or_url,
        schema_overrides=schema_overrides,
        primary_key=primary_key,
        num_retained_versions=num_retained_versions,
        comment=comment,
        extra_args=kwargs,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from IPython.lib.pretty import RepresentationPrinter


@dataclass(frozen=True)
class RowCountStats:
    """
    Statistics about the counts of rows affected by a table operation.
    """

    ins_rows: int = 0  # rows inserted
    del_rows: int = 0  # rows deleted
    upd_rows: int = 0  # rows updated
    num_excs: int = 0  # total number of exceptions
    # TODO: disambiguate what this means: # of slots computed or # of columns computed?
    computed_values: int = 0  # number of computed values (e.g., computed columns) affected by the operation

    @property
    def num_rows(self) -> int:
        return self.ins_rows + self.del_rows + self.upd_rows

    def insert_to_update(self) -> 'RowCountStats':
        """
        Convert insert row count stats to update row count stats.
        This is used when an insert operation is treated as an update.
        """
        return RowCountStats(
            ins_rows=0,
            del_rows=self.del_rows,
            upd_rows=self.upd_rows + self.ins_rows,
            num_excs=self.num_excs,
            computed_values=self.computed_values,
        )

    def __add__(self, other: 'RowCountStats') -> 'RowCountStats':
        """
        Add the stats from two RowCountStats objects together.
        """
        return RowCountStats(
            ins_rows=self.ins_rows + other.ins_rows,
            del_rows=self.del_rows + other.del_rows,
            upd_rows=self.upd_rows + other.upd_rows,
            num_excs=self.num_excs + other.num_excs,
            computed_values=self.computed_values + other.computed_values,
        )


@dataclass(frozen=True)
class UpdateStatus:
    """
    Information about changes to table data or table schema
    """

    updated_cols: list[str] = field(default_factory=list)
    cols_with_excs: list[str] = field(default_factory=list)

    # stats for the rows affected by the operation
    row_count_stats: RowCountStats = field(default_factory=RowCountStats)

    # stats for changes cascaded to other tables
    cascade_row_count_stats: RowCountStats = field(default_factory=RowCountStats)

    # stats for the rows affected by the operation in an external store
    ext_row_count_stats: RowCountStats = field(default_factory=RowCountStats)

    @property
    def num_rows(self) -> int:
        return self.row_count_stats.num_rows + self.cascade_row_count_stats.num_rows

    @property
    def num_excs(self) -> int:
        return self.row_count_stats.num_excs + self.cascade_row_count_stats.num_excs

    @property
    def num_computed_values(self) -> int:
        return self.row_count_stats.computed_values + self.cascade_row_count_stats.computed_values

    def insert_to_update(self) -> 'UpdateStatus':
        """
        Convert the update status from an insert operation to an update operation.
        This is used when an insert operation is treated as an update.
        """
        return UpdateStatus(
            updated_cols=self.updated_cols,
            cols_with_excs=self.cols_with_excs,
            row_count_stats=self.row_count_stats.insert_to_update(),
            cascade_row_count_stats=self.cascade_row_count_stats.insert_to_update(),
            ext_row_count_stats=self.ext_row_count_stats,
        )

    def to_cascade(self) -> 'UpdateStatus':
        """
        Convert the update status to a cascade update status.
        This is used when an operation cascades changes to other tables.
        """
        return UpdateStatus(
            updated_cols=self.updated_cols,
            cols_with_excs=self.cols_with_excs,
            row_count_stats=RowCountStats(),
            cascade_row_count_stats=self.cascade_row_count_stats + self.row_count_stats,
            ext_row_count_stats=self.ext_row_count_stats,
        )

    def __add__(self, other: 'UpdateStatus') -> UpdateStatus:
        """
        Add the update status from two UpdateStatus objects together.
        """
        return UpdateStatus(
            updated_cols=list(dict.fromkeys(self.updated_cols + other.updated_cols)),
            cols_with_excs=list(dict.fromkeys(self.cols_with_excs + other.cols_with_excs)),
            row_count_stats=self.row_count_stats + other.row_count_stats,
            cascade_row_count_stats=self.cascade_row_count_stats + other.cascade_row_count_stats,
            ext_row_count_stats=self.ext_row_count_stats + other.ext_row_count_stats,
        )

    @property
    def insert_msg(self) -> str:
        """Return a message describing the results of an insert operation."""
        if self.num_excs == 0:
            cols_with_excs_str = ''
        else:
            cols_with_excs_str = (
                f' across {len(self.cols_with_excs)} column{"" if len(self.cols_with_excs) == 1 else "s"}'
            )
            cols_with_excs_str += f' ({", ".join(self.cols_with_excs)})'
        msg = (
            f'Inserted {self.num_rows} row{"" if self.num_rows == 1 else "s"} '
            f'with {self.num_excs} error{"" if self.num_excs == 1 else "s"}{cols_with_excs_str}.'
        )
        return msg

    @classmethod
    def __cnt_str(cls, cnt: int, item: str) -> str:
        assert cnt > 0
        return f'{cnt} {item}{"" if cnt == 1 else "s"}'

    def _repr_pretty_(self, p: 'RepresentationPrinter', cycle: bool) -> None:
        messages = []
        # Combine row count stats and cascade row count stats
        stats = self.row_count_stats + self.cascade_row_count_stats
        if stats.ins_rows > 0:
            messages.append(f'{self.__cnt_str(stats.ins_rows, "row")} inserted')
        if stats.del_rows > 0:
            messages.append(f'{self.__cnt_str(stats.del_rows, "row")} deleted')
        if stats.upd_rows > 0:
            messages.append(f'{self.__cnt_str(stats.upd_rows, "row")} updated')
        if stats.computed_values > 0:
            messages.append(f'{self.__cnt_str(stats.computed_values, "value")} computed')
        if stats.num_excs > 0:
            messages.append(self.__cnt_str(stats.num_excs, 'exception'))
        p.text(', '.join(messages) + '.' if len(messages) > 0 else 'No rows affected.')

    @property
    def pxt_rows_updated(self) -> int:
        """
        Returns the number of Pixeltable rows that were updated as a result of the operation.
        """
        return (self.row_count_stats + self.cascade_row_count_stats).upd_rows

    @property
    def external_rows_updated(self) -> int:
        return self.ext_row_count_stats.upd_rows

    @property
    def external_rows_created(self) -> int:
        return self.ext_row_count_stats.ins_rows

    @property
    def external_rows_deleted(self) -> int:
        return self.ext_row_count_stats.del_rows

    @property
    def ext_num_rows(self) -> int:
        return self.ext_row_count_stats.num_rows



import logging
from typing import TYPE_CHECKING, Any, Optional

import pixeltable as pxt
from pixeltable import env, exprs
from pixeltable.func import Tools
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    from botocore.client import BaseClient

_logger = logging.getLogger('pixeltable')


@env.register_client('bedrock')
def _() -> 'BaseClient':
    import boto3

    return boto3.client(service_name='bedrock-runtime')


# boto3 typing is weird; type information is dynamically defined, so the best we can do for the static checker is `Any`
def _bedrock_client() -> Any:
    return env.Env.get().get_client('bedrock')


@pxt.udf
def converse(
    messages: list[dict[str, Any]],
    *,
    model_id: str,
    system: Optional[list[dict[str, Any]]] = None,
    inference_config: Optional[dict] = None,
    additional_model_request_fields: Optional[dict] = None,
    tool_config: Optional[list[dict]] = None,
) -> dict:
    """
    Generate a conversation response.

    Equivalent to the AWS Bedrock `converse` API endpoint.
    For additional details, see: <https://docs.aws.amazon.com/bedrock/latest/userguide/conversation-inference.html>

    __Requirements:__

    - `pip install boto3`

    Args:
        messages: Input messages.
        model_id: The model that will complete your prompt.
        system: An optional system prompt.
        inference_config: Base inference parameters to use.
        additional_model_request_fields: Additional inference parameters to use.

    For details on the optional parameters, see:
    <https://docs.aws.amazon.com/bedrock/latest/userguide/conversation-inference.html>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `anthropic.claude-3-haiku-20240307-v1:0`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> msgs = [{'role': 'user', 'content': [{'text': tbl.prompt}]}]
        ... tbl.add_computed_column(response=messages(msgs, model_id='anthropic.claude-3-haiku-20240307-v1:0'))
    """

    kwargs: dict[str, Any] = {'messages': messages, 'modelId': model_id}

    if system is not None:
        kwargs['system'] = system
    if inference_config is not None:
        kwargs['inferenceConfig'] = inference_config
    if additional_model_request_fields is not None:
        kwargs['additionalModelRequestFields'] = additional_model_request_fields

    if tool_config is not None:
        tool_config_ = {
            'tools': [
                {
                    'toolSpec': {
                        'name': tool['name'],
                        'description': tool['description'],
                        'inputSchema': {
                            'json': {
                                'type': 'object',
                                'properties': tool['parameters']['properties'],
                                'required': tool['required'],
                            }
                        },
                    }
                }
                for tool in tool_config
            ]
        }
        kwargs['toolConfig'] = tool_config_

    return _bedrock_client().converse(**kwargs)


def invoke_tools(tools: Tools, response: exprs.Expr) -> exprs.InlineDict:
    """Converts an Anthropic response dict to Pixeltable tool invocation format and calls `tools._invoke()`."""
    return tools._invoke(_bedrock_response_to_pxt_tool_calls(response))


@pxt.udf
def _bedrock_response_to_pxt_tool_calls(response: dict) -> Optional[dict]:
    if response.get('stopReason') != 'tool_use':
        return None

    pxt_tool_calls: dict[str, list[dict[str, Any]]] = {}
    for message in response['output']['message']['content']:
        if 'toolUse' in message:
            tool_call = message['toolUse']
            tool_name = tool_call['name']
            if tool_name not in pxt_tool_calls:
                pxt_tool_calls[tool_name] = []
            pxt_tool_calls[tool_name].append({'args': tool_call['input']})

    if len(pxt_tool_calls) == 0:
        return None

    return pxt_tool_calls


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from typing import Optional

import sqlalchemy as sql

from pixeltable import exprs


class ExecContext:
    """Class for execution runtime constants"""

    def __init__(
        self,
        row_builder: exprs.RowBuilder,
        *,
        show_pbar: bool = False,
        batch_size: int = 0,
        pk_clause: Optional[list[sql.ClauseElement]] = None,
        num_computed_exprs: int = 0,
        ignore_errors: bool = False,
    ):
        self.show_pbar = show_pbar
        self.batch_size = batch_size
        self.row_builder = row_builder
        self.profile = exprs.ExecProfile(row_builder)
        # num_rows is used to compute the total number of computed cells used for the progress bar
        self.num_rows: Optional[int] = None
        self.conn: Optional[sql.engine.Connection] = None  # if present, use this to execute SQL queries
        self.pk_clause = pk_clause
        self.num_computed_exprs = num_computed_exprs
        self.ignore_errors = ignore_errors



from typing import TYPE_CHECKING, Optional

from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    from whisperx.asr import FasterWhisperPipeline  # type: ignore[import-untyped]

import pixeltable as pxt


@pxt.udf
def transcribe(
    audio: pxt.Audio,
    *,
    model: str,
    compute_type: Optional[str] = None,
    language: Optional[str] = None,
    chunk_size: int = 30,
) -> dict:
    """
    Transcribe an audio file using WhisperX.

    This UDF runs a transcription model _locally_ using the WhisperX library,
    equivalent to the WhisperX `transcribe` function, as described in the
    [WhisperX library documentation](https://github.com/m-bain/whisperX).

    WhisperX is part of the `pixeltable.ext` package: long-term support in Pixeltable is not guaranteed.

    __Requirements:__

    - `pip install whisperx`

    Args:
        audio: The audio file to transcribe.
        model: The name of the model to use for transcription.

    See the [WhisperX library documentation](https://github.com/m-bain/whisperX) for details
    on the remaining parameters.

    Returns:
        A dictionary containing the transcription and various other metadata.

    Examples:
        Add a computed column that applies the model `tiny.en` to an existing Pixeltable column `tbl.audio`
        of the table `tbl`:

        >>> tbl.add_computed_column(result=transcribe(tbl.audio, model='tiny.en'))
    """
    import torch
    import whisperx  # type: ignore[import-untyped]

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    compute_type = compute_type or ('float16' if device == 'cuda' else 'int8')
    model = _lookup_model(model, device, compute_type)
    audio_array = whisperx.load_audio(audio)
    result = model.transcribe(audio_array, batch_size=16, language=language, chunk_size=chunk_size)
    return result


def _lookup_model(model_id: str, device: str, compute_type: str) -> 'FasterWhisperPipeline':
    import whisperx

    key = (model_id, device, compute_type)
    if key not in _model_cache:
        model = whisperx.load_model(model_id, device, compute_type=compute_type)
        _model_cache[key] = model
    return _model_cache[key]


_model_cache: dict[tuple[str, str, str], 'FasterWhisperPipeline'] = {}


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import asyncio
import inspect
from typing import TYPE_CHECKING, Any, Optional

import pixeltable as pxt
from pixeltable import exceptions as excs, type_system as ts
from pixeltable.func.signature import Parameter

if TYPE_CHECKING:
    import mcp


def mcp_udfs(url: str) -> list['pxt.func.Function']:
    return asyncio.run(mcp_udfs_async(url))


async def mcp_udfs_async(url: str) -> list['pxt.func.Function']:
    import mcp
    from mcp.client.streamable_http import streamablehttp_client

    list_tools_result: Optional[mcp.types.ListToolsResult] = None
    async with (
        streamablehttp_client(url) as (read_stream, write_stream, _),
        mcp.ClientSession(read_stream, write_stream) as session,
    ):
        await session.initialize()
        list_tools_result = await session.list_tools()
    assert list_tools_result is not None

    return [mcp_tool_to_udf(url, tool) for tool in list_tools_result.tools]


def mcp_tool_to_udf(url: str, mcp_tool: 'mcp.types.Tool') -> 'pxt.func.Function':
    import mcp
    from mcp.client.streamable_http import streamablehttp_client

    async def invoke(**kwargs: Any) -> str:
        # TODO: Cache session objects rather than creating a new one each time?
        async with (
            streamablehttp_client(url) as (read_stream, write_stream, _),
            mcp.ClientSession(read_stream, write_stream) as session,
        ):
            await session.initialize()
            res = await session.call_tool(name=mcp_tool.name, arguments=kwargs)
            # TODO Handle image/audio responses?
            return res.content[0].text  # type: ignore[union-attr]

    if mcp_tool.description is not None:
        invoke.__doc__ = mcp_tool.description

    input_schema = mcp_tool.inputSchema
    params = {
        name: __mcp_param_to_pxt_type(mcp_tool.name, name, param) for name, param in input_schema['properties'].items()
    }
    required = input_schema.get('required', [])

    # Ensure that any params not appearing in `required` are nullable.
    # (A required param might or might not be nullable, since its type might be an 'anyOf' containing a null.)
    for name in params.keys() - required:
        params[name] = params[name].copy(nullable=True)

    signature = pxt.func.Signature(
        return_type=ts.StringType(),  # Return type is always string
        parameters=[Parameter(name, col_type, inspect.Parameter.KEYWORD_ONLY) for name, col_type in params.items()],
    )

    return pxt.func.CallableFunction(signatures=[signature], py_fns=[invoke], self_name=mcp_tool.name)


def __mcp_param_to_pxt_type(tool_name: str, name: str, param: dict[str, Any]) -> ts.ColumnType:
    pxt_type = ts.ColumnType.from_json_schema(param)
    if pxt_type is None:
        raise excs.Error(f'Unknown type schema for MCP parameter {name!r} of tool {tool_name!r}: {param}')
    return pxt_type



import logging

import sqlalchemy as sql
from sqlalchemy.dialects import postgresql


def log_stmt(logger: logging.Logger, stmt: sql.sql.ClauseElement) -> None:
    logger.debug(f'executing {stmt.compile(dialect=postgresql.dialect())}')


def log_explain(logger: logging.Logger, stmt: sql.sql.ClauseElement, conn: sql.engine.Connection) -> None:
    try:
        # don't set dialect=Env.get().engine.dialect: x % y turns into x %% y, which results in a syntax error
        stmt_str = str(stmt.compile(compile_kwargs={'literal_binds': True}))
        explain_result = conn.execute(sql.text(f'EXPLAIN {stmt_str}'))
        explain_str = '\n'.join(str(row) for row in explain_result)
        logger.debug(f'SqlScanNode explain:\n{explain_str}')
    except Exception:
        logger.warning('EXPLAIN failed')



import copy
from uuid import UUID

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import (
    convert_table_md,
    convert_table_schema_version_record,
    convert_table_version_record,
)
from pixeltable.metadata.schema import TableSchemaVersion, TableVersion


@register_converter(version=30)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, table_md_updater=__update_table_md)
    convert_table_version_record(engine, table_version_record_updater=__update_table_version_record)
    convert_table_schema_version_record(
        engine, table_schema_version_record_updater=__update_table_schema_version_record
    )


def __update_table_md(md: dict, tbl_id: UUID) -> None:
    md['tbl_id'] = str(tbl_id)


def __update_table_version_record(record: TableVersion) -> None:
    """
    Update TableVersion with table_id.
    """
    assert isinstance(record.md, dict)
    md = copy.copy(record.md)
    md['tbl_id'] = str(record.tbl_id)
    record.md = md


def __update_table_schema_version_record(record: TableSchemaVersion) -> None:
    """
    Update TableSchemaVersion with table_id.
    """
    assert isinstance(record.md, dict)
    md = copy.copy(record.md)
    md['tbl_id'] = str(record.tbl_id)
    record.md = md



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `TimestampType`.

Usage example:
```python
import pixeltable as pxt

t = pxt.get_table(...)
t.select(t.timestamp_col.year, t.timestamp_col.weekday()).collect()
```
"""

from datetime import datetime
from typing import Optional

import sqlalchemy as sql

import pixeltable as pxt
from pixeltable.env import Env
from pixeltable.utils.code import local_public_names

_SQL_ZERO = sql.literal(0)


@pxt.udf(is_property=True)
def year(self: datetime) -> int:
    """
    Between [`MINYEAR`](https://docs.python.org/3/library/datetime.html#datetime.MINYEAR) and
    [`MAXYEAR`](https://docs.python.org/3/library/datetime.html#datetime.MAXYEAR) inclusive.

    Equivalent to [`datetime.year`](https://docs.python.org/3/library/datetime.html#datetime.datetime.year).
    """
    return self.year


@year.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('year', self)


@pxt.udf(is_property=True)
def month(self: datetime) -> int:
    """
    Between 1 and 12 inclusive.

    Equivalent to [`datetime.month`](https://docs.python.org/3/library/datetime.html#datetime.datetime.month).
    """
    return self.month


@month.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('month', self)


@pxt.udf(is_property=True)
def day(self: datetime) -> int:
    """
    Between 1 and the number of days in the given month of the given year.

    Equivalent to [`datetime.day`](https://docs.python.org/3/library/datetime.html#datetime.datetime.day).
    """
    return self.day


@day.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('day', self)


@pxt.udf(is_property=True)
def hour(self: datetime) -> int:
    """
    Between 0 and 23 inclusive.

    Equivalent to [`datetime.hour`](https://docs.python.org/3/library/datetime.html#datetime.datetime.hour).
    """
    return self.hour


@hour.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('hour', self)


@pxt.udf(is_property=True)
def minute(self: datetime) -> int:
    """
    Between 0 and 59 inclusive.

    Equivalent to [`datetime.minute`](https://docs.python.org/3/library/datetime.html#datetime.datetime.minute).
    """
    return self.minute


@minute.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('minute', self)


@pxt.udf(is_property=True)
def second(self: datetime) -> int:
    """
    Between 0 and 59 inclusive.

    Equivalent to [`datetime.second`](https://docs.python.org/3/library/datetime.html#datetime.datetime.second).
    """
    return self.second


@second.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('second', self)


@pxt.udf(is_property=True)
def microsecond(self: datetime) -> int:
    """
    Between 0 and 999999 inclusive.

    Equivalent to [`datetime.microsecond`](https://docs.python.org/3/library/datetime.html#datetime.datetime.microsecond).
    """
    return self.microsecond


@microsecond.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('microseconds', self) - sql.extract('second', self) * 1000000


@pxt.udf(is_method=True)
def astimezone(self: datetime, tz: str) -> datetime:
    """
    Convert the datetime to the given time zone.

    Args:
        tz: The time zone to convert to. Must be a valid time zone name from the IANA Time Zone Database.
    """
    from zoneinfo import ZoneInfo

    tzinfo = ZoneInfo(tz)
    return self.astimezone(tzinfo)


@pxt.udf(is_method=True)
def weekday(self: datetime) -> int:
    """
    Between 0 (Monday) and 6 (Sunday) inclusive.

    Equivalent to [`datetime.weekday()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.weekday).
    """
    return self.weekday()


@weekday.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('isodow', self) - 1


@pxt.udf(is_method=True)
def isoweekday(self: datetime) -> int:
    """
    Return the day of the week as an integer, where Monday is 1 and Sunday is 7.

    Equivalent to [`datetime.isoweekday()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.isoweekday).
    """
    return self.isoweekday()


@isoweekday.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('isodow', self)


@pxt.udf(is_method=True)
def isocalendar(self: datetime) -> dict:
    """
    Return a dictionary with three entries: `'year'`, `'week'`, and `'weekday'`.

    Equivalent to
    [`datetime.isocalendar()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.isocalendar).
    """
    iso_year, iso_week, iso_weekday = self.isocalendar()
    return {'year': iso_year, 'week': iso_week, 'weekday': iso_weekday}


@pxt.udf(is_method=True)
def isoformat(self: datetime, sep: str = 'T', timespec: str = 'auto') -> str:
    """
    Return a string representing the date and time in ISO 8601 format.

    Equivalent to [`datetime.isoformat()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.isoformat).

    Args:
        sep: Separator between date and time.
        timespec: The number of additional terms in the output. See the
            [`datetime.isoformat()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.isoformat)
            documentation for more details.
    """
    return self.isoformat(sep=sep, timespec=timespec)


@pxt.udf(is_method=True)
def strftime(self: datetime, format: str) -> str:
    """
    Return a string representing the date and time, controlled by an explicit format string.

    Equivalent to [`datetime.strftime()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.strftime).

    Args:
        format: The format string to control the output. For a complete list of formatting directives, see
            [`strftime()` and `strptime()` Behavior](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior).
    """
    return self.strftime(format)


@pxt.udf(is_method=True)
def make_timestamp(
    year: int, month: int, day: int, hour: int = 0, minute: int = 0, second: int = 0, microsecond: int = 0
) -> datetime:
    """
    Create a timestamp.

    Equivalent to [`datetime()`](https://docs.python.org/3/library/datetime.html#datetime.datetime).
    """
    return datetime(year, month, day, hour, minute, second, microsecond, tzinfo=Env.get().default_time_zone)


@make_timestamp.to_sql
def _(
    year: sql.ColumnElement,
    month: sql.ColumnElement,
    day: sql.ColumnElement,
    hour: sql.ColumnElement = _SQL_ZERO,
    minute: sql.ColumnElement = _SQL_ZERO,
    second: sql.ColumnElement = _SQL_ZERO,
    microsecond: sql.ColumnElement = _SQL_ZERO,
) -> sql.ColumnElement:
    return sql.func.make_timestamptz(
        sql.cast(year, sql.Integer),
        sql.cast(month, sql.Integer),
        sql.cast(day, sql.Integer),
        sql.cast(hour, sql.Integer),
        sql.cast(minute, sql.Integer),
        sql.cast(second + microsecond / 1000000.0, sql.Float),
    )


# @pxt.udf
# def date(self: datetime) -> datetime:
#     """
#     Return the date part of the datetime.
#
#     Equivalent to [`datetime.date()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.date).
#     """
#     d = self.date()
#     return datetime(d.year, d.month, d.day)
#
#
# @pxt.udf
# def time(self: datetime) -> datetime:
#     """
#     Return the time part of the datetime, with microseconds set to 0.
#
#     Equivalent to [`datetime.time()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.time).
#     """
#     t = self.time()
#     return datetime(1, 1, 1, t.hour, t.minute, t.second, t.microsecond)


@pxt.udf(is_method=True)
def replace(
    self: datetime,
    year: Optional[int] = None,
    month: Optional[int] = None,
    day: Optional[int] = None,
    hour: Optional[int] = None,
    minute: Optional[int] = None,
    second: Optional[int] = None,
    microsecond: Optional[int] = None,
) -> datetime:
    """
    Return a datetime with the same attributes, except for those attributes given new values by whichever keyword
    arguments are specified.

    Equivalent to [`datetime.replace()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.replace).
    """
    kwargs = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    return self.replace(**kwargs)


@pxt.udf(is_method=True)
def toordinal(self: datetime) -> int:
    """
    Return the proleptic Gregorian ordinal of the date, where January 1 of year 1 has ordinal 1.

    Equivalent to [`datetime.toordinal()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.toordinal).
    """
    return self.toordinal()


@pxt.udf(is_method=True)
def posix_timestamp(self: datetime) -> float:
    """
    Return POSIX timestamp corresponding to the datetime instance.

    Equivalent to [`datetime.timestamp()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.timestamp).
    """
    return self.timestamp()


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import logging
from typing import Any, AsyncIterator

from pixeltable import catalog, exprs

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode

_logger = logging.getLogger('pixeltable')


class RowUpdateNode(ExecNode):
    """
    Update individual rows in the input batches, identified by key columns.

    The updates for a row are provided as a dict of column names to new values.
    The node assumes that all update dicts contain the same keys, and it populates the slots of the columns present in
    the update list.
    """

    def __init__(
        self,
        tbl: catalog.TableVersionPath,
        key_vals_batch: list[tuple],
        is_rowid_key: bool,
        col_vals_batch: list[dict[catalog.Column, Any]],
        row_builder: exprs.RowBuilder,
        input: ExecNode,
    ):
        super().__init__(row_builder, [], [], input)
        self.updates = dict(zip(key_vals_batch, col_vals_batch))
        self.is_rowid_key = is_rowid_key
        # determine slot idxs of all columns we need to read or write
        # retrieve ColumnRefs from the RowBuilder (has slot_idx set)
        all_col_slot_idxs = {
            col_ref.col: col_ref.slot_idx
            for col_ref in row_builder.unique_exprs
            if isinstance(col_ref, exprs.ColumnRef)
        }
        self.col_slot_idxs = {col: all_col_slot_idxs[col] for col in col_vals_batch[0]}
        self.key_slot_idxs = {col: all_col_slot_idxs[col] for col in tbl.tbl_version.get().primary_key_columns()}
        self.matched_key_vals: set[tuple] = set()

    async def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        async for batch in self.input:
            for row in batch:
                key_vals = (
                    row.rowid if self.is_rowid_key else tuple(row[slot_idx] for slot_idx in self.key_slot_idxs.values())
                )
                if key_vals not in self.updates:
                    continue
                self.matched_key_vals.add(key_vals)
                col_vals = self.updates[key_vals]
                for col, val in col_vals.items():
                    slot_idx = self.col_slot_idxs[col]
                    row[slot_idx] = val
            yield batch

    def unmatched_rows(self) -> list[dict[str, Any]]:
        """Return rows that didn't get used in the updates as a list of dicts compatible with TableVersion.insert()."""
        result: list[dict[str, Any]] = []
        key_cols = self.key_slot_idxs.keys()
        for key_vals, col_vals in self.updates.items():
            if key_vals in self.matched_key_vals:
                continue
            row = {col.name: val for col, val in zip(key_cols, key_vals)}
            row.update({col.name: val for col, val in col_vals.items()})
            result.append(row)
        return result



import dataclasses
import typing
import uuid
from typing import Any, NamedTuple, Optional, TypeVar, Union, get_type_hints

import sqlalchemy as sql
from sqlalchemy import BigInteger, ForeignKey, Integer, LargeBinary, orm
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.orm.decl_api import DeclarativeMeta

from ..catalog.update_status import UpdateStatus

# Base has to be marked explicitly as a type, in order to be used elsewhere as a type hint. But in addition to being
# a type, it's also a `DeclarativeMeta`. The following pattern enables us to expose both `Base` and `Base.metadata`
# outside of the module in a typesafe way.
Base: type = orm.declarative_base()
assert isinstance(Base, DeclarativeMeta)
base_metadata = Base.metadata

T = TypeVar('T')


def md_from_dict(data_class_type: type[T], data: Any) -> T:
    """Re-instantiate a dataclass instance that contains nested dataclasses from a dict."""
    if dataclasses.is_dataclass(data_class_type):
        fieldtypes = get_type_hints(data_class_type)
        return data_class_type(**{f: md_from_dict(fieldtypes[f], data[f]) for f in data})  # type: ignore[return-value]

    origin = typing.get_origin(data_class_type)
    if origin is not None:
        type_args = typing.get_args(data_class_type)
        if origin is Union and type(None) in type_args:
            # Handling Optional types
            non_none_args = [arg for arg in type_args if arg is not type(None)]
            assert len(non_none_args) == 1
            return md_from_dict(non_none_args[0], data) if data is not None else None
        elif origin is list:
            return [md_from_dict(type_args[0], elem) for elem in data]  # type: ignore[return-value]
        elif origin is dict:
            key_type = type_args[0]
            val_type = type_args[1]
            return {key_type(key): md_from_dict(val_type, val) for key, val in data.items()}  # type: ignore[return-value]
        elif origin is tuple:
            return tuple(md_from_dict(arg_type, elem) for arg_type, elem in zip(type_args, data))  # type: ignore[return-value]
        else:
            raise AssertionError(origin)
    else:
        return data


# structure of the stored metadata:
# - each schema entity that grows somehow proportionally to the data (# of output_rows, total insert operations,
#   number of schema changes) gets its own table
# - each table has an 'md' column that basically contains the payload
# - exceptions to that are foreign keys without which lookups would be too slow (ex.: TableSchemaVersions.tbl_id)
# - the md column contains a dataclass serialized to json; this has the advantage of making changes to the metadata
#   schema easier (the goal is not to have to rely on some schema migration framework; if that breaks for some user,
#   it would be very difficult to patch up)


@dataclasses.dataclass
class SystemInfoMd:
    schema_version: int


class SystemInfo(Base):
    """A single-row table that contains system-wide metadata."""

    __tablename__ = 'systeminfo'
    dummy = sql.Column(Integer, primary_key=True, default=0, nullable=False)
    md = sql.Column(JSONB, nullable=False)  # SystemInfoMd


@dataclasses.dataclass
class DirMd:
    name: str
    user: Optional[str]
    additional_md: dict[str, Any]


class Dir(Base):
    __tablename__ = 'dirs'

    id: orm.Mapped[uuid.UUID] = orm.mapped_column(
        UUID(as_uuid=True), primary_key=True, default=uuid.uuid4, nullable=False
    )
    parent_id: orm.Mapped[uuid.UUID] = orm.mapped_column(UUID(as_uuid=True), ForeignKey('dirs.id'), nullable=True)
    md: orm.Mapped[dict[str, Any]] = orm.mapped_column(JSONB, nullable=False)  # DirMd

    # used to force acquisition of an X-lock via an Update stmt
    lock_dummy: orm.Mapped[int] = orm.mapped_column(BigInteger, nullable=True)


@dataclasses.dataclass
class ColumnMd:
    """
    Records the non-versioned metadata of a column.
    - immutable attributes: type, primary key, etc.
    - when a column was added/dropped, which is needed to GC unreachable storage columns
      (a column that was added after table snapshot n and dropped before table snapshot n+1 can be removed
      from the stored table).
    """

    id: int
    schema_version_add: int
    schema_version_drop: Optional[int]
    col_type: dict

    # if True, is part of the primary key
    is_pk: bool

    # if set, this is a computed column
    value_expr: Optional[dict]

    # if True, the column is present in the stored table
    stored: Optional[bool]


@dataclasses.dataclass
class IndexMd:
    """
    Metadata needed to instantiate an EmbeddingIndex
    """

    id: int
    name: str
    indexed_col_tbl_id: str  # UUID of the table (as string) that contains column being indexed
    indexed_col_id: int  # column being indexed
    index_val_col_id: int  # column holding the values to be indexed
    index_val_undo_col_id: int  # column holding index values for deleted rows
    schema_version_add: int
    schema_version_drop: Optional[int]
    class_fqn: str
    init_args: dict[str, Any]


# a stored table version path is a list of (table id as str, effective table version)
TableVersionPath = list[tuple[str, Optional[int]]]


@dataclasses.dataclass
class ViewMd:
    is_snapshot: bool
    include_base_columns: bool

    # (table id, version); for mutable views, all versions are None
    base_versions: TableVersionPath

    # filter predicate applied to the base table; view-only
    predicate: Optional[dict[str, Any]]

    # sampling predicate applied to the base table; view-only
    sample_clause: Optional[dict[str, Any]]

    # ComponentIterator subclass; only for component views
    iterator_class_fqn: Optional[str]

    # args to pass to the iterator class constructor; only for component views
    iterator_args: Optional[dict[str, Any]]


@dataclasses.dataclass
class TableMd:
    tbl_id: str  # uuid.UUID
    name: str
    is_replica: bool

    user: Optional[str]

    # monotonically increasing w/in Table for both data and schema changes, starting at 0
    current_version: int
    # each version has a corresponding schema version (current_version >= current_schema_version)
    current_schema_version: int

    next_col_id: int  # used to assign Column.id
    next_idx_id: int  # used to assign IndexMd.id

    # - used to assign the rowid column in the storage table
    # - every row is assigned a unique and immutable rowid on insertion
    next_row_id: int

    # sequence number to track changes in the set of mutable views of this table (ie, this table = the view base)
    # - incremented for each add/drop of a mutable view
    # - only maintained for mutable tables
    # TODO: replace with mutable_views: list[UUID] to help with debugging
    view_sn: int

    # Metadata format for external stores:
    # {'class': 'pixeltable.io.label_studio.LabelStudioProject', 'md': {'project_id': 3}}
    external_stores: list[dict[str, Any]]

    column_md: dict[int, ColumnMd]  # col_id -> ColumnMd
    index_md: dict[int, IndexMd]  # index_id -> IndexMd
    view_md: Optional[ViewMd]
    additional_md: dict[str, Any]

    has_pending_ops: bool = False

    @property
    def is_snapshot(self) -> bool:
        return self.view_md is not None and self.view_md.is_snapshot

    @property
    def is_mutable(self) -> bool:
        return not self.is_snapshot and not self.is_replica

    @property
    def is_pure_snapshot(self) -> bool:
        return (
            self.view_md is not None
            and self.view_md.is_snapshot
            and self.view_md.sample_clause is None
            and self.view_md.predicate is None
            and len(self.column_md) == 0
        )


class Table(Base):
    """
    Table represents both tables and views.

    Views are in essence a subclass of tables, because they also store materialized columns. The differences are:
    - views have a base, which is either a (live) table or a snapshot
    - views can have a filter predicate
    """

    __tablename__ = 'tables'

    MAX_VERSION = 9223372036854775807  # 2^63 - 1

    id: orm.Mapped[uuid.UUID] = orm.mapped_column(UUID(as_uuid=True), primary_key=True, nullable=False)
    dir_id: orm.Mapped[uuid.UUID] = orm.mapped_column(UUID(as_uuid=True), ForeignKey('dirs.id'), nullable=False)
    md: orm.Mapped[dict[str, Any]] = orm.mapped_column(JSONB, nullable=False)  # TableMd

    # used to force acquisition of an X-lock via an Update stmt
    lock_dummy: orm.Mapped[int] = orm.mapped_column(BigInteger, nullable=True)


@dataclasses.dataclass
class TableVersionMd:
    tbl_id: str  # uuid.UUID
    created_at: float  # time.time()
    version: int
    schema_version: int
    user: Optional[str] = None  # User that created this version
    update_status: Optional[UpdateStatus] = None  # UpdateStatus of the change that created this version
    additional_md: dict[str, Any] = dataclasses.field(default_factory=dict)


class TableVersion(Base):
    __tablename__ = 'tableversions'
    tbl_id: orm.Mapped[uuid.UUID] = orm.mapped_column(
        UUID(as_uuid=True), ForeignKey('tables.id'), primary_key=True, nullable=False
    )
    version: orm.Mapped[int] = orm.mapped_column(BigInteger, primary_key=True, nullable=False)
    md: orm.Mapped[dict[str, Any]] = orm.mapped_column(JSONB, nullable=False)


@dataclasses.dataclass
class SchemaColumn:
    """
    Records the versioned metadata of a column.
    """

    pos: int
    name: str

    # media validation strategy of this particular media column; if not set, TableMd.media_validation applies
    # stores column.MediaValiation.name.lower()
    media_validation: Optional[str]


@dataclasses.dataclass
class TableSchemaVersionMd:
    """
    Records all versioned table metadata.
    """

    tbl_id: str  # uuid.UUID
    schema_version: int
    preceding_schema_version: Optional[int]
    columns: dict[int, SchemaColumn]  # col_id -> SchemaColumn
    num_retained_versions: int
    comment: str

    # default validation strategy for any media column of this table
    # stores column.MediaValiation.name.lower()
    media_validation: str
    additional_md: dict[str, Any]


# versioning: each table schema change results in a new record
class TableSchemaVersion(Base):
    __tablename__ = 'tableschemaversions'

    tbl_id: orm.Mapped[uuid.UUID] = orm.mapped_column(
        UUID(as_uuid=True), ForeignKey('tables.id'), primary_key=True, nullable=False
    )
    schema_version: orm.Mapped[int] = orm.mapped_column(BigInteger, primary_key=True, nullable=False)
    md: orm.Mapped[dict[str, Any]] = orm.mapped_column(JSONB, nullable=False)  # TableSchemaVersionMd


class PendingTableOp(Base):
    """
    Table operation that needs to be completed before the table can be used.

    Operations need to be completed in order of increasing seq_num.
    """

    __tablename__ = 'pendingtableops'

    tbl_id: orm.Mapped[uuid.UUID] = orm.mapped_column(
        UUID(as_uuid=True), ForeignKey('tables.id'), primary_key=True, nullable=False
    )
    op_sn: orm.Mapped[int] = orm.mapped_column(Integer, primary_key=True, nullable=False)  # catalog.TableOp.op_sn
    op: orm.Mapped[dict[str, Any]] = orm.mapped_column(JSONB, nullable=False)  # catalog.TableOp


@dataclasses.dataclass
class FunctionMd:
    name: str
    py_version: str  # platform.python_version
    class_name: str  # name of the Function subclass
    md: dict  # part of the output of Function.to_store()


class Function(Base):
    """
    User-defined functions that are not module functions (ie, aren't available at runtime as a symbol in a known
    module).
    Functions without a name are anonymous functions used in the definition of a computed column.
    Functions that have names are also assigned to a database and directory.
    We store the Python version under which a Function was created (and the callable pickled) in order to warn
    against version mismatches.
    """

    __tablename__ = 'functions'

    id: orm.Mapped[uuid.UUID] = orm.mapped_column(
        UUID(as_uuid=True), primary_key=True, default=uuid.uuid4, nullable=False
    )
    dir_id: orm.Mapped[uuid.UUID] = orm.mapped_column(UUID(as_uuid=True), ForeignKey('dirs.id'), nullable=True)
    md: orm.Mapped[dict[str, Any]] = orm.mapped_column(JSONB, nullable=False)  # FunctionMd
    binary_obj: orm.Mapped[Optional[bytes]] = orm.mapped_column(LargeBinary, nullable=True)


class FullTableMd(NamedTuple):
    tbl_md: TableMd
    version_md: TableVersionMd
    schema_version_md: TableSchemaVersionMd

    @property
    def is_pure_snapshot(self) -> bool:
        return (
            self.tbl_md.view_md is not None
            and self.tbl_md.view_md.predicate is None
            and len(self.schema_version_md.columns) == 0
        )

    def as_dict(self) -> dict[str, Any]:
        return {
            'table_id': self.tbl_md.tbl_id,
            'table_md': dataclasses.asdict(self.tbl_md),
            'table_version_md': dataclasses.asdict(self.version_md),
            'table_schema_version_md': dataclasses.asdict(self.schema_version_md),
        }

    @classmethod
    def from_dict(cls, data_dict: dict[str, Any]) -> 'FullTableMd':
        return FullTableMd(
            tbl_md=md_from_dict(TableMd, data_dict['table_md']),
            version_md=md_from_dict(TableVersionMd, data_dict['table_version_md']),
            schema_version_md=md_from_dict(TableSchemaVersionMd, data_dict['table_schema_version_md']),
        )



import sqlalchemy as sql

from pixeltable.metadata import register_converter


@register_converter(version=31)
def _(engine: sql.engine.Engine) -> None:
    # Add a column "lock_dummy: int8" to the dirs table in the store
    # This column is the target of an UPDATE operation to synchronize directory operations
    with engine.begin() as conn:
        conn.execute(sql.text('ALTER TABLE dirs ADD COLUMN lock_dummy int8'))
